#include "rogue.h"

#include "nir.h"
#include "nir_builder.h"
#include "nir_search.h"
#include "nir_search_helpers.h"

/* What follows is NIR algebraic transform code for the following 46
 * transforms:
 *    ('extract_u16', 'a@32', 'b@32') => ('ubitfield_extract', 'a', ('imul', 'b', 16), 16)
 *    ('extract_i16', 'a@32', 'b@32') => ('ibitfield_extract', 'a', ('imul', 'b', 16), 16)
 *    ('extract_u8', 'a@32', 'b@32') => ('ubitfield_extract', 'a', ('imul', 'b', 8), 8)
 *    ('extract_i8', 'a@32', 'b@32') => ('ibitfield_extract', 'a', ('imul', 'b', 8), 8)
 *    ('bitfield_insert', 'base', 'insert', 'offset', 0) => base
 *    ('ubitfield_extract', 'value', 'offset', 0) => 0
 *    ('ibitfield_extract', 'value', 'offset', 0) => 0
 *    ('iadd@64', 'a', '#offset(is_upper_half_zero)') => ('uadd64_32_img', 'a', ('u2u32', 'offset'))
 *    ('unpack_snorm_2x16', 'a') => ('vec2', ('unpack_snorm_2x16_split_x', 'a'), ('unpack_snorm_2x16_split_y', 'a'))
 *    ('unpack_snorm_4x8', 'a') => ('vec4', ('unpack_snorm_4x8_split_x', 'a'), ('unpack_snorm_4x8_split_y', 'a'), ('unpack_snorm_4x8_split_z', 'a'), ('unpack_snorm_4x8_split_w', 'a'))
 *    ('unpack_snorm_r10g10b10a2', 'a') => ('vec4', ('unpack_snorm_r10g10b10a2_split_x', 'a'), ('unpack_snorm_r10g10b10a2_split_y', 'a'), ('unpack_snorm_r10g10b10a2_split_z', 'a'), ('unpack_snorm_r10g10b10a2_split_w', 'a'))
 *    ('unpack_unorm_2x16', 'a') => ('vec2', ('unpack_unorm_2x16_split_x', 'a'), ('unpack_unorm_2x16_split_y', 'a'))
 *    ('unpack_unorm_4x8', 'a') => ('vec4', ('unpack_unorm_4x8_split_x', 'a'), ('unpack_unorm_4x8_split_y', 'a'), ('unpack_unorm_4x8_split_z', 'a'), ('unpack_unorm_4x8_split_w', 'a'))
 *    ('unpack_unorm_r10g10b10a2', 'a') => ('vec4', ('unpack_unorm_r10g10b10a2_split_x', 'a'), ('unpack_unorm_r10g10b10a2_split_y', 'a'), ('unpack_unorm_r10g10b10a2_split_z', 'a'), ('unpack_unorm_r10g10b10a2_split_w', 'a'))
 *    ('unpack_sscaled_2x16', 'a') => ('vec2', ('unpack_sscaled_2x16_split_x', 'a'), ('unpack_sscaled_2x16_split_y', 'a'))
 *    ('unpack_sscaled_4x8', 'a') => ('vec4', ('unpack_sscaled_4x8_split_x', 'a'), ('unpack_sscaled_4x8_split_y', 'a'), ('unpack_sscaled_4x8_split_z', 'a'), ('unpack_sscaled_4x8_split_w', 'a'))
 *    ('unpack_sscaled_r10g10b10a2', 'a') => ('vec4', ('unpack_sscaled_r10g10b10a2_split_x', 'a'), ('unpack_sscaled_r10g10b10a2_split_y', 'a'), ('unpack_sscaled_r10g10b10a2_split_z', 'a'), ('unpack_sscaled_r10g10b10a2_split_w', 'a'))
 *    ('unpack_uscaled_2x16', 'a') => ('vec2', ('unpack_uscaled_2x16_split_x', 'a'), ('unpack_uscaled_2x16_split_y', 'a'))
 *    ('unpack_uscaled_4x8', 'a') => ('vec4', ('unpack_uscaled_4x8_split_x', 'a'), ('unpack_uscaled_4x8_split_y', 'a'), ('unpack_uscaled_4x8_split_z', 'a'), ('unpack_uscaled_4x8_split_w', 'a'))
 *    ('unpack_half_2x16', 'a') => ('vec2', ('unpack_half_2x16_split_x', 'a'), ('unpack_half_2x16_split_y', 'a'))
 *    ('unpack_r11g11b10f', 'a') => ('vec3', ('unpack_r11g11b10f_split_x', 'a'), ('unpack_r11g11b10f_split_y', 'a'), ('unpack_r11g11b10f_split_z', 'a'))
 *    ('pack_snorm_2x16', 'a') => ('pack_snorm_2x16_field', ('pack_snorm_2x16_field', 0, 'a.x', 0), 'a.y', 1)
 *    ('pack_snorm_4x8', 'a') => ('pack_snorm_4x8_field', ('pack_snorm_4x8_field', ('pack_snorm_4x8_field', ('pack_snorm_4x8_field', 0, 'a.x', 0), 'a.y', 1), 'a.z', 2), 'a.w', 3)
 *    ('pack_unorm_2x16', 'a') => ('pack_unorm_2x16_field', ('pack_unorm_2x16_field', 0, 'a.x', 0), 'a.y', 1)
 *    ('pack_unorm_4x8', 'a') => ('pack_unorm_4x8_field', ('pack_unorm_4x8_field', ('pack_unorm_4x8_field', ('pack_unorm_4x8_field', 0, 'a.x', 0), 'a.y', 1), 'a.z', 2), 'a.w', 3)
 *    ('pack_sscaled_2x16', 'a') => ('pack_sscaled_2x16_field', ('pack_sscaled_2x16_field', 0, 'a.x', 0), 'a.y', 1)
 *    ('pack_sscaled_4x8', 'a') => ('pack_sscaled_4x8_field', ('pack_sscaled_4x8_field', ('pack_sscaled_4x8_field', ('pack_sscaled_4x8_field', 0, 'a.x', 0), 'a.y', 1), 'a.z', 2), 'a.w', 3)
 *    ('pack_uscaled_2x16', 'a') => ('pack_uscaled_2x16_field', ('pack_uscaled_2x16_field', 0, 'a.x', 0), 'a.y', 1)
 *    ('pack_uscaled_4x8', 'a') => ('pack_uscaled_4x8_field', ('pack_uscaled_4x8_field', ('pack_uscaled_4x8_field', ('pack_uscaled_4x8_field', 0, 'a.x', 0), 'a.y', 1), 'a.z', 2), 'a.w', 3)
 *    ('pack_half_2x16', 'a') => ('pack_half_2x16_field', ('pack_half_2x16_field', 0, 'a.x', 0), 'a.y', 1)
 *    ('pack_half_2x16_split', 'a', 'b') => ('pack_half_2x16_field', ('pack_half_2x16_field', 0, 'a', 0), 'b', 1)
 *    ('pack_snorm_r10g10b10a2_field', 'base', 'insert', 0) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_snorm_r10g10b10a2', ('vec4', 'insert', 0, 0, 0)), 0, 10), 0, 10)
 *    ('pack_snorm_r10g10b10a2_field', 'base', 'insert', 1) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_snorm_r10g10b10a2', ('vec4', 0, 'insert', 0, 0)), 10, 10), 10, 10)
 *    ('pack_snorm_r10g10b10a2_field', 'base', 'insert', 2) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_snorm_r10g10b10a2', ('vec4', 0, 0, 'insert', 0)), 20, 10), 20, 10)
 *    ('pack_snorm_r10g10b10a2_field', 'base', 'insert', 3) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_snorm_r10g10b10a2', ('vec4', 0, 0, 0, 'insert')), 30, 2), 30, 2)
 *    ('pack_unorm_r10g10b10a2_field', 'base', 'insert', 0) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_unorm_r10g10b10a2', ('vec4', 'insert', 0, 0, 0)), 0, 10), 0, 10)
 *    ('pack_unorm_r10g10b10a2_field', 'base', 'insert', 1) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_unorm_r10g10b10a2', ('vec4', 0, 'insert', 0, 0)), 10, 10), 10, 10)
 *    ('pack_unorm_r10g10b10a2_field', 'base', 'insert', 2) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_unorm_r10g10b10a2', ('vec4', 0, 0, 'insert', 0)), 20, 10), 20, 10)
 *    ('pack_unorm_r10g10b10a2_field', 'base', 'insert', 3) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_unorm_r10g10b10a2', ('vec4', 0, 0, 0, 'insert')), 30, 2), 30, 2)
 *    ('pack_sscaled_r10g10b10a2_field', 'base', 'insert', 0) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_sscaled_r10g10b10a2', ('vec4', 'insert', 0, 0, 0)), 0, 10), 0, 10)
 *    ('pack_sscaled_r10g10b10a2_field', 'base', 'insert', 1) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_sscaled_r10g10b10a2', ('vec4', 0, 'insert', 0, 0)), 10, 10), 10, 10)
 *    ('pack_sscaled_r10g10b10a2_field', 'base', 'insert', 2) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_sscaled_r10g10b10a2', ('vec4', 0, 0, 'insert', 0)), 20, 10), 20, 10)
 *    ('pack_sscaled_r10g10b10a2_field', 'base', 'insert', 3) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_sscaled_r10g10b10a2', ('vec4', 0, 0, 0, 'insert')), 30, 2), 30, 2)
 *    ('pack_r11g11b10f_field', 'base', 'insert', 0) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_r11g11b10f', ('vec3', 'insert', 0, 0)), 0, 11), 0, 11)
 *    ('pack_r11g11b10f_field', 'base', 'insert', 1) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_r11g11b10f', ('vec3', 0, 'insert', 0)), 11, 11), 11, 11)
 *    ('pack_r11g11b10f_field', 'base', 'insert', 2) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_r11g11b10f', ('vec3', 0, 0, 'insert')), 22, 10), 22, 10)
 */


static const nir_search_value_union rogue_nir_opt_algebraic_late_values[] = {
   /* ('extract_u16', 'a@32', 'b@32') => ('ubitfield_extract', 'a', ('imul', 'b', 16), 16) */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace0_0 -> 0 in the cache */
   /* replace0_1_0 -> 1 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x10 /* 16 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 1, 3 },
      -1,
   } },
   /* replace0_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 1,
      { 0, 4, 3 },
      -1,
   } },

   /* ('extract_i16', 'a@32', 'b@32') => ('ibitfield_extract', 'a', ('imul', 'b', 16), 16) */
   /* search1_0 -> 0 in the cache */
   /* search1_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace1_0 -> 0 in the cache */
   /* replace1_1_0 -> 1 in the cache */
   /* replace1_1_1 -> 3 in the cache */
   /* replace1_1 -> 4 in the cache */
   /* replace1_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibitfield_extract,
      -1, 1,
      { 0, 4, 3 },
      -1,
   } },

   /* ('extract_u8', 'a@32', 'b@32') => ('ubitfield_extract', 'a', ('imul', 'b', 8), 8) */
   /* search2_0 -> 0 in the cache */
   /* search2_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace2_0 -> 0 in the cache */
   /* replace2_1_0 -> 1 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x8 /* 8 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 1, 9 },
      -1,
   } },
   /* replace2_2 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 1,
      { 0, 10, 9 },
      -1,
   } },

   /* ('extract_i8', 'a@32', 'b@32') => ('ibitfield_extract', 'a', ('imul', 'b', 8), 8) */
   /* search3_0 -> 0 in the cache */
   /* search3_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace3_0 -> 0 in the cache */
   /* replace3_1_0 -> 1 in the cache */
   /* replace3_1_1 -> 9 in the cache */
   /* replace3_1 -> 10 in the cache */
   /* replace3_2 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibitfield_extract,
      -1, 1,
      { 0, 10, 9 },
      -1,
   } },

   /* ('bitfield_insert', 'base', 'insert', 'offset', 0) => base */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* base */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* insert */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* offset */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_insert,
      -1, 0,
      { 14, 15, 16, 17 },
      -1,
   } },

   /* replace4 -> 14 in the cache */

   /* ('ubitfield_extract', 'value', 'offset', 0) => 0 */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* value */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* offset */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search5_2 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 0,
      { 19, 20, 17 },
      -1,
   } },

   /* replace5 -> 17 in the cache */

   /* ('ibitfield_extract', 'value', 'offset', 0) => 0 */
   /* search6_0 -> 19 in the cache */
   /* search6_1 -> 20 in the cache */
   /* search6_2 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibitfield_extract,
      -1, 0,
      { 19, 20, 17 },
      -1,
   } },

   /* replace6 -> 17 in the cache */

   /* ('iadd@64', 'a', '#offset(is_upper_half_zero)') => ('uadd64_32_img', 'a', ('u2u32', 'offset')) */
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* offset */
      true,
      nir_type_invalid,
      0,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 23, 24 },
      -1,
   } },

   /* replace7_0 -> 23 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* offset */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 26 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_uadd64_32_img,
      0, 1,
      { 23, 27 },
      -1,
   } },

   /* ('unpack_snorm_2x16', 'a') => ('vec2', ('unpack_snorm_2x16_split_x', 'a'), ('unpack_snorm_2x16_split_y', 'a')) */
   /* search8_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_snorm_2x16,
      -1, 0,
      { 0 },
      -1,
   } },

   /* replace8_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_snorm_2x16_split_x,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace8_1_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_snorm_2x16_split_y,
      -1, 0,
      { 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 30, 31 },
      -1,
   } },

   /* ('unpack_snorm_4x8', 'a') => ('vec4', ('unpack_snorm_4x8_split_x', 'a'), ('unpack_snorm_4x8_split_y', 'a'), ('unpack_snorm_4x8_split_z', 'a'), ('unpack_snorm_4x8_split_w', 'a')) */
   /* search9_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_snorm_4x8,
      -1, 0,
      { 0 },
      -1,
   } },

   /* replace9_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_snorm_4x8_split_x,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace9_1_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_snorm_4x8_split_y,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace9_2_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_snorm_4x8_split_z,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace9_3_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_snorm_4x8_split_w,
      -1, 0,
      { 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 34, 35, 36, 37 },
      -1,
   } },

   /* ('unpack_snorm_r10g10b10a2', 'a') => ('vec4', ('unpack_snorm_r10g10b10a2_split_x', 'a'), ('unpack_snorm_r10g10b10a2_split_y', 'a'), ('unpack_snorm_r10g10b10a2_split_z', 'a'), ('unpack_snorm_r10g10b10a2_split_w', 'a')) */
   /* search10_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_snorm_r10g10b10a2,
      -1, 0,
      { 0 },
      -1,
   } },

   /* replace10_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_snorm_r10g10b10a2_split_x,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace10_1_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_snorm_r10g10b10a2_split_y,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace10_2_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_snorm_r10g10b10a2_split_z,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace10_3_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_snorm_r10g10b10a2_split_w,
      -1, 0,
      { 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 40, 41, 42, 43 },
      -1,
   } },

   /* ('unpack_unorm_2x16', 'a') => ('vec2', ('unpack_unorm_2x16_split_x', 'a'), ('unpack_unorm_2x16_split_y', 'a')) */
   /* search11_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_unorm_2x16,
      -1, 0,
      { 0 },
      -1,
   } },

   /* replace11_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_unorm_2x16_split_x,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace11_1_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_unorm_2x16_split_y,
      -1, 0,
      { 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 46, 47 },
      -1,
   } },

   /* ('unpack_unorm_4x8', 'a') => ('vec4', ('unpack_unorm_4x8_split_x', 'a'), ('unpack_unorm_4x8_split_y', 'a'), ('unpack_unorm_4x8_split_z', 'a'), ('unpack_unorm_4x8_split_w', 'a')) */
   /* search12_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_unorm_4x8,
      -1, 0,
      { 0 },
      -1,
   } },

   /* replace12_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_unorm_4x8_split_x,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace12_1_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_unorm_4x8_split_y,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace12_2_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_unorm_4x8_split_z,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace12_3_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_unorm_4x8_split_w,
      -1, 0,
      { 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 50, 51, 52, 53 },
      -1,
   } },

   /* ('unpack_unorm_r10g10b10a2', 'a') => ('vec4', ('unpack_unorm_r10g10b10a2_split_x', 'a'), ('unpack_unorm_r10g10b10a2_split_y', 'a'), ('unpack_unorm_r10g10b10a2_split_z', 'a'), ('unpack_unorm_r10g10b10a2_split_w', 'a')) */
   /* search13_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_unorm_r10g10b10a2,
      -1, 0,
      { 0 },
      -1,
   } },

   /* replace13_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_unorm_r10g10b10a2_split_x,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace13_1_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_unorm_r10g10b10a2_split_y,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace13_2_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_unorm_r10g10b10a2_split_z,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace13_3_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_unorm_r10g10b10a2_split_w,
      -1, 0,
      { 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 56, 57, 58, 59 },
      -1,
   } },

   /* ('unpack_sscaled_2x16', 'a') => ('vec2', ('unpack_sscaled_2x16_split_x', 'a'), ('unpack_sscaled_2x16_split_y', 'a')) */
   /* search14_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_sscaled_2x16,
      -1, 0,
      { 0 },
      -1,
   } },

   /* replace14_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_sscaled_2x16_split_x,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace14_1_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_sscaled_2x16_split_y,
      -1, 0,
      { 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 62, 63 },
      -1,
   } },

   /* ('unpack_sscaled_4x8', 'a') => ('vec4', ('unpack_sscaled_4x8_split_x', 'a'), ('unpack_sscaled_4x8_split_y', 'a'), ('unpack_sscaled_4x8_split_z', 'a'), ('unpack_sscaled_4x8_split_w', 'a')) */
   /* search15_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_sscaled_4x8,
      -1, 0,
      { 0 },
      -1,
   } },

   /* replace15_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_sscaled_4x8_split_x,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace15_1_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_sscaled_4x8_split_y,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace15_2_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_sscaled_4x8_split_z,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace15_3_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_sscaled_4x8_split_w,
      -1, 0,
      { 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 66, 67, 68, 69 },
      -1,
   } },

   /* ('unpack_sscaled_r10g10b10a2', 'a') => ('vec4', ('unpack_sscaled_r10g10b10a2_split_x', 'a'), ('unpack_sscaled_r10g10b10a2_split_y', 'a'), ('unpack_sscaled_r10g10b10a2_split_z', 'a'), ('unpack_sscaled_r10g10b10a2_split_w', 'a')) */
   /* search16_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_sscaled_r10g10b10a2,
      -1, 0,
      { 0 },
      -1,
   } },

   /* replace16_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_sscaled_r10g10b10a2_split_x,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace16_1_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_sscaled_r10g10b10a2_split_y,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace16_2_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_sscaled_r10g10b10a2_split_z,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace16_3_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_sscaled_r10g10b10a2_split_w,
      -1, 0,
      { 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 72, 73, 74, 75 },
      -1,
   } },

   /* ('unpack_uscaled_2x16', 'a') => ('vec2', ('unpack_uscaled_2x16_split_x', 'a'), ('unpack_uscaled_2x16_split_y', 'a')) */
   /* search17_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_uscaled_2x16,
      -1, 0,
      { 0 },
      -1,
   } },

   /* replace17_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_uscaled_2x16_split_x,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace17_1_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_uscaled_2x16_split_y,
      -1, 0,
      { 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 78, 79 },
      -1,
   } },

   /* ('unpack_uscaled_4x8', 'a') => ('vec4', ('unpack_uscaled_4x8_split_x', 'a'), ('unpack_uscaled_4x8_split_y', 'a'), ('unpack_uscaled_4x8_split_z', 'a'), ('unpack_uscaled_4x8_split_w', 'a')) */
   /* search18_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_uscaled_4x8,
      -1, 0,
      { 0 },
      -1,
   } },

   /* replace18_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_uscaled_4x8_split_x,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace18_1_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_uscaled_4x8_split_y,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace18_2_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_uscaled_4x8_split_z,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace18_3_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_uscaled_4x8_split_w,
      -1, 0,
      { 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 82, 83, 84, 85 },
      -1,
   } },

   /* ('unpack_half_2x16', 'a') => ('vec2', ('unpack_half_2x16_split_x', 'a'), ('unpack_half_2x16_split_y', 'a')) */
   /* search19_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16,
      -1, 0,
      { 0 },
      -1,
   } },

   /* replace19_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_x,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace19_1_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_y,
      -1, 0,
      { 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 88, 89 },
      -1,
   } },

   /* ('unpack_r11g11b10f', 'a') => ('vec3', ('unpack_r11g11b10f_split_x', 'a'), ('unpack_r11g11b10f_split_y', 'a'), ('unpack_r11g11b10f_split_z', 'a')) */
   /* search20_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_r11g11b10f,
      -1, 0,
      { 0 },
      -1,
   } },

   /* replace20_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_r11g11b10f_split_x,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace20_1_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_r11g11b10f_split_y,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace20_2_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_r11g11b10f_split_z,
      -1, 0,
      { 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 92, 93, 94 },
      -1,
   } },

   /* ('pack_snorm_2x16', 'a') => ('pack_snorm_2x16_field', ('pack_snorm_2x16_field', 0, 'a.x', 0), 'a.y', 1) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_2x16,
      -1, 0,
      { 96 },
      -1,
   } },

   /* replace21_0_0 -> 17 in the cache */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0},
   } },
   /* replace21_0_2 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_2x16_field,
      -1, 0,
      { 17, 98, 17 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {1},
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_2x16_field,
      -1, 0,
      { 99, 100, 101 },
      -1,
   } },

   /* ('pack_snorm_4x8', 'a') => ('pack_snorm_4x8_field', ('pack_snorm_4x8_field', ('pack_snorm_4x8_field', ('pack_snorm_4x8_field', 0, 'a.x', 0), 'a.y', 1), 'a.z', 2), 'a.w', 3) */
   /* search22_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_4x8,
      -1, 0,
      { 0 },
      -1,
   } },

   /* replace22_0_0_0_0 -> 17 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0},
   } },
   /* replace22_0_0_0_2 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_4x8_field,
      -1, 0,
      { 17, 104, 17 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {1},
   } },
   /* replace22_0_0_2 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_4x8_field,
      -1, 0,
      { 105, 106, 101 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {2},
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x2 /* 2 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_4x8_field,
      -1, 0,
      { 107, 108, 109 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {3},
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3 /* 3 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_4x8_field,
      -1, 0,
      { 110, 111, 112 },
      -1,
   } },

   /* ('pack_unorm_2x16', 'a') => ('pack_unorm_2x16_field', ('pack_unorm_2x16_field', 0, 'a.x', 0), 'a.y', 1) */
   /* search23_0 -> 96 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_2x16,
      -1, 0,
      { 96 },
      -1,
   } },

   /* replace23_0_0 -> 17 in the cache */
   /* replace23_0_1 -> 98 in the cache */
   /* replace23_0_2 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_2x16_field,
      -1, 0,
      { 17, 98, 17 },
      -1,
   } },
   /* replace23_1 -> 100 in the cache */
   /* replace23_2 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_2x16_field,
      -1, 0,
      { 115, 100, 101 },
      -1,
   } },

   /* ('pack_unorm_4x8', 'a') => ('pack_unorm_4x8_field', ('pack_unorm_4x8_field', ('pack_unorm_4x8_field', ('pack_unorm_4x8_field', 0, 'a.x', 0), 'a.y', 1), 'a.z', 2), 'a.w', 3) */
   /* search24_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_4x8,
      -1, 0,
      { 0 },
      -1,
   } },

   /* replace24_0_0_0_0 -> 17 in the cache */
   /* replace24_0_0_0_1 -> 104 in the cache */
   /* replace24_0_0_0_2 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_4x8_field,
      -1, 0,
      { 17, 104, 17 },
      -1,
   } },
   /* replace24_0_0_1 -> 106 in the cache */
   /* replace24_0_0_2 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_4x8_field,
      -1, 0,
      { 118, 106, 101 },
      -1,
   } },
   /* replace24_0_1 -> 108 in the cache */
   /* replace24_0_2 -> 109 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_4x8_field,
      -1, 0,
      { 119, 108, 109 },
      -1,
   } },
   /* replace24_1 -> 111 in the cache */
   /* replace24_2 -> 112 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_4x8_field,
      -1, 0,
      { 120, 111, 112 },
      -1,
   } },

   /* ('pack_sscaled_2x16', 'a') => ('pack_sscaled_2x16_field', ('pack_sscaled_2x16_field', 0, 'a.x', 0), 'a.y', 1) */
   /* search25_0 -> 96 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_sscaled_2x16,
      -1, 0,
      { 96 },
      -1,
   } },

   /* replace25_0_0 -> 17 in the cache */
   /* replace25_0_1 -> 98 in the cache */
   /* replace25_0_2 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_sscaled_2x16_field,
      -1, 0,
      { 17, 98, 17 },
      -1,
   } },
   /* replace25_1 -> 100 in the cache */
   /* replace25_2 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_sscaled_2x16_field,
      -1, 0,
      { 123, 100, 101 },
      -1,
   } },

   /* ('pack_sscaled_4x8', 'a') => ('pack_sscaled_4x8_field', ('pack_sscaled_4x8_field', ('pack_sscaled_4x8_field', ('pack_sscaled_4x8_field', 0, 'a.x', 0), 'a.y', 1), 'a.z', 2), 'a.w', 3) */
   /* search26_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_sscaled_4x8,
      -1, 0,
      { 0 },
      -1,
   } },

   /* replace26_0_0_0_0 -> 17 in the cache */
   /* replace26_0_0_0_1 -> 104 in the cache */
   /* replace26_0_0_0_2 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_sscaled_4x8_field,
      -1, 0,
      { 17, 104, 17 },
      -1,
   } },
   /* replace26_0_0_1 -> 106 in the cache */
   /* replace26_0_0_2 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_sscaled_4x8_field,
      -1, 0,
      { 126, 106, 101 },
      -1,
   } },
   /* replace26_0_1 -> 108 in the cache */
   /* replace26_0_2 -> 109 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_sscaled_4x8_field,
      -1, 0,
      { 127, 108, 109 },
      -1,
   } },
   /* replace26_1 -> 111 in the cache */
   /* replace26_2 -> 112 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_sscaled_4x8_field,
      -1, 0,
      { 128, 111, 112 },
      -1,
   } },

   /* ('pack_uscaled_2x16', 'a') => ('pack_uscaled_2x16_field', ('pack_uscaled_2x16_field', 0, 'a.x', 0), 'a.y', 1) */
   /* search27_0 -> 96 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uscaled_2x16,
      -1, 0,
      { 96 },
      -1,
   } },

   /* replace27_0_0 -> 17 in the cache */
   /* replace27_0_1 -> 98 in the cache */
   /* replace27_0_2 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uscaled_2x16_field,
      -1, 0,
      { 17, 98, 17 },
      -1,
   } },
   /* replace27_1 -> 100 in the cache */
   /* replace27_2 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uscaled_2x16_field,
      -1, 0,
      { 131, 100, 101 },
      -1,
   } },

   /* ('pack_uscaled_4x8', 'a') => ('pack_uscaled_4x8_field', ('pack_uscaled_4x8_field', ('pack_uscaled_4x8_field', ('pack_uscaled_4x8_field', 0, 'a.x', 0), 'a.y', 1), 'a.z', 2), 'a.w', 3) */
   /* search28_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uscaled_4x8,
      -1, 0,
      { 0 },
      -1,
   } },

   /* replace28_0_0_0_0 -> 17 in the cache */
   /* replace28_0_0_0_1 -> 104 in the cache */
   /* replace28_0_0_0_2 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uscaled_4x8_field,
      -1, 0,
      { 17, 104, 17 },
      -1,
   } },
   /* replace28_0_0_1 -> 106 in the cache */
   /* replace28_0_0_2 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uscaled_4x8_field,
      -1, 0,
      { 134, 106, 101 },
      -1,
   } },
   /* replace28_0_1 -> 108 in the cache */
   /* replace28_0_2 -> 109 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uscaled_4x8_field,
      -1, 0,
      { 135, 108, 109 },
      -1,
   } },
   /* replace28_1 -> 111 in the cache */
   /* replace28_2 -> 112 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uscaled_4x8_field,
      -1, 0,
      { 136, 111, 112 },
      -1,
   } },

   /* ('pack_half_2x16', 'a') => ('pack_half_2x16_field', ('pack_half_2x16_field', 0, 'a.x', 0), 'a.y', 1) */
   /* search29_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16,
      -1, 0,
      { 0 },
      -1,
   } },

   /* replace29_0_0 -> 17 in the cache */
   /* replace29_0_1 -> 104 in the cache */
   /* replace29_0_2 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_field,
      -1, 0,
      { 17, 104, 17 },
      -1,
   } },
   /* replace29_1 -> 106 in the cache */
   /* replace29_2 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_field,
      -1, 0,
      { 139, 106, 101 },
      -1,
   } },

   /* ('pack_half_2x16_split', 'a', 'b') => ('pack_half_2x16_field', ('pack_half_2x16_field', 0, 'a', 0), 'b', 1) */
   /* search30_0 -> 0 in the cache */
   /* search30_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_split,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace30_0_0 -> 17 in the cache */
   /* replace30_0_1 -> 0 in the cache */
   /* replace30_0_2 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_field,
      -1, 0,
      { 17, 0, 17 },
      -1,
   } },
   /* replace30_1 -> 1 in the cache */
   /* replace30_2 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_field,
      -1, 0,
      { 142, 1, 101 },
      -1,
   } },

   /* ('pack_snorm_r10g10b10a2_field', 'base', 'insert', 0) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_snorm_r10g10b10a2', ('vec4', 'insert', 0, 0, 0)), 0, 10), 0, 10) */
   /* search31_0 -> 14 in the cache */
   /* search31_1 -> 15 in the cache */
   /* search31_2 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_r10g10b10a2_field,
      -1, 0,
      { 14, 15, 17 },
      -1,
   } },

   /* replace31_0 -> 14 in the cache */
   /* replace31_1_0_0_0 -> 15 in the cache */
   /* replace31_1_0_0_1 -> 17 in the cache */
   /* replace31_1_0_0_2 -> 17 in the cache */
   /* replace31_1_0_0_3 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 15, 17, 17, 17 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_r10g10b10a2,
      -1, 0,
      { 145 },
      -1,
   } },
   /* replace31_1_1 -> 17 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xa /* 10 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 0,
      { 146, 17, 147 },
      -1,
   } },
   /* replace31_2 -> 17 in the cache */
   /* replace31_3 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_insert,
      -1, 0,
      { 14, 148, 17, 147 },
      -1,
   } },

   /* ('pack_snorm_r10g10b10a2_field', 'base', 'insert', 1) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_snorm_r10g10b10a2', ('vec4', 0, 'insert', 0, 0)), 10, 10), 10, 10) */
   /* search32_0 -> 14 in the cache */
   /* search32_1 -> 15 in the cache */
   /* search32_2 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_r10g10b10a2_field,
      -1, 0,
      { 14, 15, 101 },
      -1,
   } },

   /* replace32_0 -> 14 in the cache */
   /* replace32_1_0_0_0 -> 17 in the cache */
   /* replace32_1_0_0_1 -> 15 in the cache */
   /* replace32_1_0_0_2 -> 17 in the cache */
   /* replace32_1_0_0_3 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 17, 15, 17, 17 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_r10g10b10a2,
      -1, 0,
      { 151 },
      -1,
   } },
   /* replace32_1_1 -> 147 in the cache */
   /* replace32_1_2 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 0,
      { 152, 147, 147 },
      -1,
   } },
   /* replace32_2 -> 147 in the cache */
   /* replace32_3 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_insert,
      -1, 0,
      { 14, 153, 147, 147 },
      -1,
   } },

   /* ('pack_snorm_r10g10b10a2_field', 'base', 'insert', 2) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_snorm_r10g10b10a2', ('vec4', 0, 0, 'insert', 0)), 20, 10), 20, 10) */
   /* search33_0 -> 14 in the cache */
   /* search33_1 -> 15 in the cache */
   /* search33_2 -> 109 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_r10g10b10a2_field,
      -1, 0,
      { 14, 15, 109 },
      -1,
   } },

   /* replace33_0 -> 14 in the cache */
   /* replace33_1_0_0_0 -> 17 in the cache */
   /* replace33_1_0_0_1 -> 17 in the cache */
   /* replace33_1_0_0_2 -> 15 in the cache */
   /* replace33_1_0_0_3 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 17, 17, 15, 17 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_r10g10b10a2,
      -1, 0,
      { 156 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x14 /* 20 */ },
   } },
   /* replace33_1_2 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 0,
      { 157, 158, 147 },
      -1,
   } },
   /* replace33_2 -> 158 in the cache */
   /* replace33_3 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_insert,
      -1, 0,
      { 14, 159, 158, 147 },
      -1,
   } },

   /* ('pack_snorm_r10g10b10a2_field', 'base', 'insert', 3) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_snorm_r10g10b10a2', ('vec4', 0, 0, 0, 'insert')), 30, 2), 30, 2) */
   /* search34_0 -> 14 in the cache */
   /* search34_1 -> 15 in the cache */
   /* search34_2 -> 112 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_r10g10b10a2_field,
      -1, 0,
      { 14, 15, 112 },
      -1,
   } },

   /* replace34_0 -> 14 in the cache */
   /* replace34_1_0_0_0 -> 17 in the cache */
   /* replace34_1_0_0_1 -> 17 in the cache */
   /* replace34_1_0_0_2 -> 17 in the cache */
   /* replace34_1_0_0_3 -> 15 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 17, 17, 17, 15 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_r10g10b10a2,
      -1, 0,
      { 162 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1e /* 30 */ },
   } },
   /* replace34_1_2 -> 109 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 0,
      { 163, 164, 109 },
      -1,
   } },
   /* replace34_2 -> 164 in the cache */
   /* replace34_3 -> 109 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_insert,
      -1, 0,
      { 14, 165, 164, 109 },
      -1,
   } },

   /* ('pack_unorm_r10g10b10a2_field', 'base', 'insert', 0) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_unorm_r10g10b10a2', ('vec4', 'insert', 0, 0, 0)), 0, 10), 0, 10) */
   /* search35_0 -> 14 in the cache */
   /* search35_1 -> 15 in the cache */
   /* search35_2 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_r10g10b10a2_field,
      -1, 0,
      { 14, 15, 17 },
      -1,
   } },

   /* replace35_0 -> 14 in the cache */
   /* replace35_1_0_0_0 -> 15 in the cache */
   /* replace35_1_0_0_1 -> 17 in the cache */
   /* replace35_1_0_0_2 -> 17 in the cache */
   /* replace35_1_0_0_3 -> 17 in the cache */
   /* replace35_1_0_0 -> 145 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_r10g10b10a2,
      -1, 0,
      { 145 },
      -1,
   } },
   /* replace35_1_1 -> 17 in the cache */
   /* replace35_1_2 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 0,
      { 168, 17, 147 },
      -1,
   } },
   /* replace35_2 -> 17 in the cache */
   /* replace35_3 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_insert,
      -1, 0,
      { 14, 169, 17, 147 },
      -1,
   } },

   /* ('pack_unorm_r10g10b10a2_field', 'base', 'insert', 1) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_unorm_r10g10b10a2', ('vec4', 0, 'insert', 0, 0)), 10, 10), 10, 10) */
   /* search36_0 -> 14 in the cache */
   /* search36_1 -> 15 in the cache */
   /* search36_2 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_r10g10b10a2_field,
      -1, 0,
      { 14, 15, 101 },
      -1,
   } },

   /* replace36_0 -> 14 in the cache */
   /* replace36_1_0_0_0 -> 17 in the cache */
   /* replace36_1_0_0_1 -> 15 in the cache */
   /* replace36_1_0_0_2 -> 17 in the cache */
   /* replace36_1_0_0_3 -> 17 in the cache */
   /* replace36_1_0_0 -> 151 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_r10g10b10a2,
      -1, 0,
      { 151 },
      -1,
   } },
   /* replace36_1_1 -> 147 in the cache */
   /* replace36_1_2 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 0,
      { 172, 147, 147 },
      -1,
   } },
   /* replace36_2 -> 147 in the cache */
   /* replace36_3 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_insert,
      -1, 0,
      { 14, 173, 147, 147 },
      -1,
   } },

   /* ('pack_unorm_r10g10b10a2_field', 'base', 'insert', 2) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_unorm_r10g10b10a2', ('vec4', 0, 0, 'insert', 0)), 20, 10), 20, 10) */
   /* search37_0 -> 14 in the cache */
   /* search37_1 -> 15 in the cache */
   /* search37_2 -> 109 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_r10g10b10a2_field,
      -1, 0,
      { 14, 15, 109 },
      -1,
   } },

   /* replace37_0 -> 14 in the cache */
   /* replace37_1_0_0_0 -> 17 in the cache */
   /* replace37_1_0_0_1 -> 17 in the cache */
   /* replace37_1_0_0_2 -> 15 in the cache */
   /* replace37_1_0_0_3 -> 17 in the cache */
   /* replace37_1_0_0 -> 156 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_r10g10b10a2,
      -1, 0,
      { 156 },
      -1,
   } },
   /* replace37_1_1 -> 158 in the cache */
   /* replace37_1_2 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 0,
      { 176, 158, 147 },
      -1,
   } },
   /* replace37_2 -> 158 in the cache */
   /* replace37_3 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_insert,
      -1, 0,
      { 14, 177, 158, 147 },
      -1,
   } },

   /* ('pack_unorm_r10g10b10a2_field', 'base', 'insert', 3) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_unorm_r10g10b10a2', ('vec4', 0, 0, 0, 'insert')), 30, 2), 30, 2) */
   /* search38_0 -> 14 in the cache */
   /* search38_1 -> 15 in the cache */
   /* search38_2 -> 112 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_r10g10b10a2_field,
      -1, 0,
      { 14, 15, 112 },
      -1,
   } },

   /* replace38_0 -> 14 in the cache */
   /* replace38_1_0_0_0 -> 17 in the cache */
   /* replace38_1_0_0_1 -> 17 in the cache */
   /* replace38_1_0_0_2 -> 17 in the cache */
   /* replace38_1_0_0_3 -> 15 in the cache */
   /* replace38_1_0_0 -> 162 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_r10g10b10a2,
      -1, 0,
      { 162 },
      -1,
   } },
   /* replace38_1_1 -> 164 in the cache */
   /* replace38_1_2 -> 109 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 0,
      { 180, 164, 109 },
      -1,
   } },
   /* replace38_2 -> 164 in the cache */
   /* replace38_3 -> 109 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_insert,
      -1, 0,
      { 14, 181, 164, 109 },
      -1,
   } },

   /* ('pack_sscaled_r10g10b10a2_field', 'base', 'insert', 0) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_sscaled_r10g10b10a2', ('vec4', 'insert', 0, 0, 0)), 0, 10), 0, 10) */
   /* search39_0 -> 14 in the cache */
   /* search39_1 -> 15 in the cache */
   /* search39_2 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_sscaled_r10g10b10a2_field,
      -1, 0,
      { 14, 15, 17 },
      -1,
   } },

   /* replace39_0 -> 14 in the cache */
   /* replace39_1_0_0_0 -> 15 in the cache */
   /* replace39_1_0_0_1 -> 17 in the cache */
   /* replace39_1_0_0_2 -> 17 in the cache */
   /* replace39_1_0_0_3 -> 17 in the cache */
   /* replace39_1_0_0 -> 145 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_sscaled_r10g10b10a2,
      -1, 0,
      { 145 },
      -1,
   } },
   /* replace39_1_1 -> 17 in the cache */
   /* replace39_1_2 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 0,
      { 184, 17, 147 },
      -1,
   } },
   /* replace39_2 -> 17 in the cache */
   /* replace39_3 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_insert,
      -1, 0,
      { 14, 185, 17, 147 },
      -1,
   } },

   /* ('pack_sscaled_r10g10b10a2_field', 'base', 'insert', 1) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_sscaled_r10g10b10a2', ('vec4', 0, 'insert', 0, 0)), 10, 10), 10, 10) */
   /* search40_0 -> 14 in the cache */
   /* search40_1 -> 15 in the cache */
   /* search40_2 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_sscaled_r10g10b10a2_field,
      -1, 0,
      { 14, 15, 101 },
      -1,
   } },

   /* replace40_0 -> 14 in the cache */
   /* replace40_1_0_0_0 -> 17 in the cache */
   /* replace40_1_0_0_1 -> 15 in the cache */
   /* replace40_1_0_0_2 -> 17 in the cache */
   /* replace40_1_0_0_3 -> 17 in the cache */
   /* replace40_1_0_0 -> 151 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_sscaled_r10g10b10a2,
      -1, 0,
      { 151 },
      -1,
   } },
   /* replace40_1_1 -> 147 in the cache */
   /* replace40_1_2 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 0,
      { 188, 147, 147 },
      -1,
   } },
   /* replace40_2 -> 147 in the cache */
   /* replace40_3 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_insert,
      -1, 0,
      { 14, 189, 147, 147 },
      -1,
   } },

   /* ('pack_sscaled_r10g10b10a2_field', 'base', 'insert', 2) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_sscaled_r10g10b10a2', ('vec4', 0, 0, 'insert', 0)), 20, 10), 20, 10) */
   /* search41_0 -> 14 in the cache */
   /* search41_1 -> 15 in the cache */
   /* search41_2 -> 109 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_sscaled_r10g10b10a2_field,
      -1, 0,
      { 14, 15, 109 },
      -1,
   } },

   /* replace41_0 -> 14 in the cache */
   /* replace41_1_0_0_0 -> 17 in the cache */
   /* replace41_1_0_0_1 -> 17 in the cache */
   /* replace41_1_0_0_2 -> 15 in the cache */
   /* replace41_1_0_0_3 -> 17 in the cache */
   /* replace41_1_0_0 -> 156 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_sscaled_r10g10b10a2,
      -1, 0,
      { 156 },
      -1,
   } },
   /* replace41_1_1 -> 158 in the cache */
   /* replace41_1_2 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 0,
      { 192, 158, 147 },
      -1,
   } },
   /* replace41_2 -> 158 in the cache */
   /* replace41_3 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_insert,
      -1, 0,
      { 14, 193, 158, 147 },
      -1,
   } },

   /* ('pack_sscaled_r10g10b10a2_field', 'base', 'insert', 3) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_sscaled_r10g10b10a2', ('vec4', 0, 0, 0, 'insert')), 30, 2), 30, 2) */
   /* search42_0 -> 14 in the cache */
   /* search42_1 -> 15 in the cache */
   /* search42_2 -> 112 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_sscaled_r10g10b10a2_field,
      -1, 0,
      { 14, 15, 112 },
      -1,
   } },

   /* replace42_0 -> 14 in the cache */
   /* replace42_1_0_0_0 -> 17 in the cache */
   /* replace42_1_0_0_1 -> 17 in the cache */
   /* replace42_1_0_0_2 -> 17 in the cache */
   /* replace42_1_0_0_3 -> 15 in the cache */
   /* replace42_1_0_0 -> 162 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_sscaled_r10g10b10a2,
      -1, 0,
      { 162 },
      -1,
   } },
   /* replace42_1_1 -> 164 in the cache */
   /* replace42_1_2 -> 109 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 0,
      { 196, 164, 109 },
      -1,
   } },
   /* replace42_2 -> 164 in the cache */
   /* replace42_3 -> 109 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_insert,
      -1, 0,
      { 14, 197, 164, 109 },
      -1,
   } },

   /* ('pack_r11g11b10f_field', 'base', 'insert', 0) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_r11g11b10f', ('vec3', 'insert', 0, 0)), 0, 11), 0, 11) */
   /* search43_0 -> 14 in the cache */
   /* search43_1 -> 15 in the cache */
   /* search43_2 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_r11g11b10f_field,
      -1, 0,
      { 14, 15, 17 },
      -1,
   } },

   /* replace43_0 -> 14 in the cache */
   /* replace43_1_0_0_0 -> 15 in the cache */
   /* replace43_1_0_0_1 -> 17 in the cache */
   /* replace43_1_0_0_2 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 15, 17, 17 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_r11g11b10f,
      -1, 0,
      { 200 },
      -1,
   } },
   /* replace43_1_1 -> 17 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xb /* 11 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 0,
      { 201, 17, 202 },
      -1,
   } },
   /* replace43_2 -> 17 in the cache */
   /* replace43_3 -> 202 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_insert,
      -1, 0,
      { 14, 203, 17, 202 },
      -1,
   } },

   /* ('pack_r11g11b10f_field', 'base', 'insert', 1) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_r11g11b10f', ('vec3', 0, 'insert', 0)), 11, 11), 11, 11) */
   /* search44_0 -> 14 in the cache */
   /* search44_1 -> 15 in the cache */
   /* search44_2 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_r11g11b10f_field,
      -1, 0,
      { 14, 15, 101 },
      -1,
   } },

   /* replace44_0 -> 14 in the cache */
   /* replace44_1_0_0_0 -> 17 in the cache */
   /* replace44_1_0_0_1 -> 15 in the cache */
   /* replace44_1_0_0_2 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 17, 15, 17 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_r11g11b10f,
      -1, 0,
      { 206 },
      -1,
   } },
   /* replace44_1_1 -> 202 in the cache */
   /* replace44_1_2 -> 202 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 0,
      { 207, 202, 202 },
      -1,
   } },
   /* replace44_2 -> 202 in the cache */
   /* replace44_3 -> 202 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_insert,
      -1, 0,
      { 14, 208, 202, 202 },
      -1,
   } },

   /* ('pack_r11g11b10f_field', 'base', 'insert', 2) => ('bitfield_insert', 'base', ('ubitfield_extract', ('pack_r11g11b10f', ('vec3', 0, 0, 'insert')), 22, 10), 22, 10) */
   /* search45_0 -> 14 in the cache */
   /* search45_1 -> 15 in the cache */
   /* search45_2 -> 109 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_r11g11b10f_field,
      -1, 0,
      { 14, 15, 109 },
      -1,
   } },

   /* replace45_0 -> 14 in the cache */
   /* replace45_1_0_0_0 -> 17 in the cache */
   /* replace45_1_0_0_1 -> 17 in the cache */
   /* replace45_1_0_0_2 -> 15 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 17, 17, 15 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_r11g11b10f,
      -1, 0,
      { 211 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x16 /* 22 */ },
   } },
   /* replace45_1_2 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 0,
      { 212, 213, 147 },
      -1,
   } },
   /* replace45_2 -> 213 in the cache */
   /* replace45_3 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_insert,
      -1, 0,
      { 14, 214, 213, 147 },
      -1,
   } },

};


static const nir_search_variable_cond rogue_nir_opt_algebraic_late_variable_cond[] = {
   (is_upper_half_zero),
};

static const struct transform rogue_nir_opt_algebraic_late_transforms[] = {
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 6, 7, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 8, 11, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 12, 13, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 18, 14, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 21, 17, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 22, 17, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 25, 28, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 29, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 33, 38, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 39, 44, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 45, 48, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 49, 54, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 55, 60, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 61, 64, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 70, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 71, 76, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 77, 80, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 81, 86, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 87, 90, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 91, 95, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 97, 102, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 103, 113, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 116, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 117, 121, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 122, 124, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 125, 129, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 130, 132, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 133, 137, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 138, 140, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 141, 143, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 144, 149, 0 },
   { 150, 154, 0 },
   { 155, 160, 0 },
   { 161, 166, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 167, 170, 0 },
   { 171, 174, 0 },
   { 175, 178, 0 },
   { 179, 182, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 183, 186, 0 },
   { 187, 190, 0 },
   { 191, 194, 0 },
   { 195, 198, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 199, 204, 0 },
   { 205, 209, 0 },
   { 210, 215, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

};

static const struct per_op_table rogue_nir_opt_algebraic_late_pass_op_table[nir_num_search_ops] = {
   [nir_op_extract_u16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         2,
      },
   },
   [nir_op_extract_i16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         3,
      },
   },
   [nir_op_extract_u8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         4,
      },
   },
   [nir_op_extract_i8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         5,
      },
   },
   [nir_op_bitfield_insert] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         6,
         0,
         6,
         0,
         6,
         0,
         6,
         0,
         6,
         0,
         6,
         0,
         6,
         0,
         6,
      },
   },
   [nir_op_ubitfield_extract] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         7,
         0,
         7,
         0,
         7,
         0,
         7,
      },
   },
   [nir_op_ibitfield_extract] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         8,
         0,
         8,
         0,
         8,
         0,
         8,
      },
   },
   [nir_op_iadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         9,
         9,
         9,
      },
   },
   [nir_op_unpack_snorm_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         10,
      },
   },
   [nir_op_unpack_snorm_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         11,
      },
   },
   [nir_op_unpack_snorm_r10g10b10a2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         12,
      },
   },
   [nir_op_unpack_unorm_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         13,
      },
   },
   [nir_op_unpack_unorm_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         14,
      },
   },
   [nir_op_unpack_unorm_r10g10b10a2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         15,
      },
   },
   [nir_op_unpack_sscaled_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         16,
      },
   },
   [nir_op_unpack_sscaled_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         17,
      },
   },
   [nir_op_unpack_sscaled_r10g10b10a2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         18,
      },
   },
   [nir_op_unpack_uscaled_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         19,
      },
   },
   [nir_op_unpack_uscaled_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         20,
      },
   },
   [nir_op_unpack_half_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         21,
      },
   },
   [nir_op_unpack_r11g11b10f] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         22,
      },
   },
   [nir_op_pack_snorm_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         23,
      },
   },
   [nir_op_pack_snorm_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         24,
      },
   },
   [nir_op_pack_unorm_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         25,
      },
   },
   [nir_op_pack_unorm_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         26,
      },
   },
   [nir_op_pack_sscaled_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         27,
      },
   },
   [nir_op_pack_sscaled_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         28,
      },
   },
   [nir_op_pack_uscaled_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         29,
      },
   },
   [nir_op_pack_uscaled_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         30,
      },
   },
   [nir_op_pack_half_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         31,
      },
   },
   [nir_op_pack_half_2x16_split] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         32,
      },
   },
   [nir_op_pack_snorm_r10g10b10a2_field] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         33,
         0,
         33,
         0,
         33,
         0,
         33,
      },
   },
   [nir_op_pack_unorm_r10g10b10a2_field] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         34,
         0,
         34,
         0,
         34,
         0,
         34,
      },
   },
   [nir_op_pack_sscaled_r10g10b10a2_field] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         35,
         0,
         35,
         0,
         35,
         0,
         35,
      },
   },
   [nir_op_pack_r11g11b10f_field] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         36,
         0,
         36,
         0,
         36,
         0,
         36,
      },
   },
};

/* Mapping from state index to offset in transforms (0 being no transforms) */
static const uint16_t rogue_nir_opt_algebraic_late_transform_offsets[] = {
   0,
   0,
   1,
   3,
   5,
   7,
   9,
   11,
   13,
   15,
   17,
   19,
   21,
   23,
   25,
   27,
   29,
   31,
   33,
   35,
   37,
   39,
   41,
   43,
   45,
   47,
   49,
   51,
   53,
   55,
   57,
   59,
   61,
   63,
   68,
   73,
   78,
};

static const nir_algebraic_table rogue_nir_opt_algebraic_late_table = {
   .transforms = rogue_nir_opt_algebraic_late_transforms,
   .transform_offsets = rogue_nir_opt_algebraic_late_transform_offsets,
   .pass_op_table = rogue_nir_opt_algebraic_late_pass_op_table,
   .values = rogue_nir_opt_algebraic_late_values,
   .expression_cond = NULL,
   .variable_cond = rogue_nir_opt_algebraic_late_variable_cond,
};

bool
rogue_nir_opt_algebraic_late(
   nir_shader *shader
) {
   bool progress = false;
   bool condition_flags[1];
   const nir_shader_compiler_options *options = shader->options;
   const shader_info *info = &shader->info;
   (void) options;
   (void) info;

   STATIC_ASSERT(216 == ARRAY_SIZE(rogue_nir_opt_algebraic_late_values));
   condition_flags[0] = true;

   nir_foreach_function_impl(impl, shader) {
     progress |= nir_algebraic_impl(impl, condition_flags, &rogue_nir_opt_algebraic_late_table);
   }

   return progress;
}

