/*
 * Copyright © 2024 Imagination Technologies Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef PVR_PDS_ISA_H
#define PVR_PDS_ISA_H

#include <stdbool.h>
#include <stdint.h>

#include "util/macros.h"

/*****************************************************************************
 * Hardware sensitive enums.
 *****************************************************************************
 * These enums match the values understood by the hardware.
 */

enum pvr_pds_predicate {
   PVR_PDS_PREDICATE_P0,
   PVR_PDS_PREDICATE_IF0,
   PVR_PDS_PREDICATE_IF1,
   PVR_PDS_PREDICATE_SO_OVERFLOW_PREDICATE_0,
   PVR_PDS_PREDICATE_SO_OVERFLOW_PREDICATE_1,
   PVR_PDS_PREDICATE_SO_OVERFLOW_PREDICATE_2,
   PVR_PDS_PREDICATE_SO_OVERFLOW_PREDICATE_3,
   PVR_PDS_PREDICATE_SO_OVERFLOW_PREDICATE_GLOBAL,
   PVR_PDS_PREDICATE_KEEP,
};

/**
 * SOMASK bits for clearing a specific predicate with STMC.
 */
enum pvr_pds_so_mask_bit {
   PVR_PDS_SO_MASK_STREAM_0 = BITFIELD_BIT(0),
   PVR_PDS_SO_MASK_STREAM_1 = BITFIELD_BIT(1),
   PVR_PDS_SO_MASK_STREAM_2 = BITFIELD_BIT(2),
   PVR_PDS_SO_MASK_STREAM_3 = BITFIELD_BIT(3),
   PVR_PDS_SO_MASK_GLOBAL = BITFIELD_BIT(4),
};

enum pvr_pds_logic_op_type {
   PVR_PDS_LOGIC_OP_NONE,
   PVR_PDS_LOGIC_OP_NOT,
   PVR_PDS_LOGIC_OP_AND,
   PVR_PDS_LOGIC_OP_OR,
   PVR_PDS_LOGIC_OP_XOR,
   PVR_PDS_LOGIC_OP_XNOR,
   PVR_PDS_LOGIC_OP_NAND,
   PVR_PDS_LOGIC_OP_NOR,
};

enum pvr_pds_comparison_op_type {
   PVR_PDS_LOGIC_OP_EQ,
   PVR_PDS_LOGIC_OP_GT,
   PVR_PDS_LOGIC_OP_LT,
   PVR_PDS_LOGIC_OP_NE,
};

enum pvr_pds_dout_cmd_type {
   PVR_PDS_DOUT_CMD_DOUTD,
   PVR_PDS_DOUT_CMD_DOUTW,
   PVR_PDS_DOUT_CMD_DOUTU,
   PVR_PDS_DOUT_CMD_DOUTV,
   PVR_PDS_DOUT_CMD_DOUTI,
   PVR_PDS_DOUT_CMD_DOUTC,
};

/* Temporary registers that get filled by DMs before executing specific PDS
 * tasks.
 */
enum pvr_pds_prefilled_temp {
   PVR_PDS_PREFILLED_TEMP_VS_VERTEX_ID = 0,
   PVR_PDS_PREFILLED_TEMP_VS_INSTANCE_ID = 1,

   /* Only available in the coefficient sync task. */
   PVR_PDS_PREFILLED_TEMP_CDM_WORK_GROUP_ID_X = 0,
   PVR_PDS_PREFILLED_TEMP_CDM_WORK_GROUP_ID_Y = 1,
   PVR_PDS_PREFILLED_TEMP_CDM_WORK_GROUP_ID_Z = 2,
   /* Available in every other CDM related task. */
   PVR_PDS_PREFILLED_TEMP_CDM_LOCAL_ID_X = 0,
   PVR_PDS_PREFILLED_TEMP_CDM_LOCAL_ID_YZ = 1,

   PVR_PDS_PREFILLED_TEMP_GS_INSTANCE_ID = 0,
   PVR_PDS_PREFILLED_TEMP_GS_PRIMITIVE_ID = 1,
   PVR_PDS_PREFILLED_TEMP_GS_INDECIES_0_3 = 2,
   PVR_PDS_PREFILLED_TEMP_GS_INDECIES_4_7 = 3,
   PVR_PDS_PREFILLED_TEMP_GS_INDECIES_8_11 = 4,
   PVR_PDS_PREFILLED_TEMP_GS_INDECIES_12_15 = 5,
   PVR_PDS_PREFILLED_TEMP_GS_INDECIES_16_19 = 6,
   PVR_PDS_PREFILLED_TEMP_GS_INDECIES_20_23 = 7,
   PVR_PDS_PREFILLED_TEMP_GS_INDECIES_24_27 = 8,
   PVR_PDS_PREFILLED_TEMP_GS_INDECIES_28_31 = 9,
};

/*****************************************************************************
 * PDS Register.
 *****************************************************************************/

#define PVR_PDS_REG_CONST_COUNT 128
#define PVR_PDS_REG_TEMP_COUNT 32
#define PVR_PDS_REG_PTEMP_COUNT 32

enum pvr_pds_reg_type {
   PVR_PDS_REG_TYPE_INVALID = 0,
   PVR_PDS_REG_TYPE_CONST,
   PVR_PDS_REG_TYPE_TEMP,
   PVR_PDS_REG_TYPE_PTEMP,

   PVR_PDS_REG_TYPE_COUNT,
};

enum pvr_pds_reg_size {
   PVR_PDS_REG_SIZE_INVALID = 0,
   PVR_PDS_REG_SIZE_32,
   PVR_PDS_REG_SIZE_64,

   PVR_PDS_REG_SIZE_COUNT,
};

struct pvr_pds_reg {
   enum pvr_pds_reg_type type : 8;
   enum pvr_pds_reg_size reg_size : 8;

   uint32_t idx : 8;
};

static_assert(PVR_PDS_REG_TYPE_COUNT <= (1 << 8),
              "'struct pvr_pds_reg' 'type' too small");
static_assert(PVR_PDS_REG_SIZE_COUNT <= (1 << 8),
              "'struct pvr_pds_reg' 'reg_size' too small");

static_assert(PVR_PDS_REG_CONST_COUNT <= (1 << 8),
              "'struct pvr_pds_reg' 'idx' too small");
static_assert(PVR_PDS_REG_TEMP_COUNT <= (1 << 8),
              "'struct pvr_pds_reg' 'idx' too small");
static_assert(PVR_PDS_REG_PTEMP_COUNT <= (1 << 8),
              "'struct pvr_pds_reg' 'idx' too small");

/*****************************************************************************
 * PDS Instruction.
 *****************************************************************************/

/* SO instruction takes in the most amount of sources, 4 sources + a stream
 * index immediate so a total of 5 sources.
 *
 * Everything else takes in less than 4 sources, so including the instruction
 * destination for them we always end up with 5 as the maximum.
 */
#define PVR_PDS_MAX_SRC_COUNT (5)

/* The destination register is always the last element, if the instruction has
 * a destination.
 */
#define PVR_PDS_INSTR_DST (PVR_PDS_MAX_SRC_COUNT - 1)

enum pvr_pds_op_type {
   PVR_PDS_OP_NOP,

   PVR_PDS_OP_MAD,
   PVR_PDS_OP_ADD32,
   PVR_PDS_OP_ADD64,
   PVR_PDS_OP_SFTLP,
   PVR_PDS_OP_SFTLP32,
   PVR_PDS_OP_SFTLP64,

   PVR_PDS_OP_CMP,
   PVR_PDS_OP_BRA,
   PVR_PDS_OP_LD,
   PVR_PDS_OP_LIMM,
   PVR_PDS_OP_ST,

   PVR_PDS_OP_STM,
   PVR_PDS_OP_STMC,
   PVR_PDS_OP_DDMAD,
   PVR_PDS_OP_DOUT,
   PVR_PDS_OP_WDF,
   PVR_PDS_OP_LOCK,
   PVR_PDS_OP_RELEASE,
   PVR_PDS_OP_HALT,
};

enum pvr_pds_instr_flag_bit {
   /* MAD, ADD32, ADD64 - Subtract instead of adding.
    * BRA - Invert condition (the meaning of the predicate is inverted).
    */
   PVR_PDS_INSTR_FLAG_INVERT = BITFIELD_BIT(0),
   /* MAD, ADD32, ADD64 - Perform signed math. */
   PVR_PDS_INSTR_FLAG_SIGNED_OP = BITFIELD_BIT(1),
   /* STM - Set the predicates but do not write data to memory. */
   PVR_PDS_INSTR_FLAG_TST = BITFIELD_BIT(2),
   /* DDMAD, DOUT - Halt execution after this instruction. */
   PVR_PDS_INSTR_FLAG_END = BITFIELD_BIT(3),
   /* LIMM - immediate source is actually a global register index. */
   PVR_PDS_INSTR_FLAG_IMM_IS_GR = BITFIELD_BIT(4),
};

enum pvr_pds_instr_src_type {
   PVR_PDS_SRC_TYPE_INVALID = 0,
   PVR_PDS_SRC_TYPE_REGISTER,
   PVR_PDS_SRC_TYPE_IMMEDIATE,
};

struct pvr_pds_instr_src {
   enum pvr_pds_instr_src_type type;

   union {
      uint32_t imm;
      struct pvr_pds_reg reg;
   };
};

struct pvr_pds_instr {
   enum pvr_pds_op_type type;

   union {
      /* Only valid for SFTLP `type`. */
      enum pvr_pds_logic_op_type logic;

      /* Only valid for CMP `type`. */
      enum pvr_pds_comparison_op_type comparison;

      /* Only valid for DOUT `type`. */
      enum pvr_pds_dout_cmd_type dout_cmd;
   } sub_op_type;

   /* Mask composed of `enum pvr_pds_instr_flag_bit`. */
   uint32_t flags;

   struct {
      /* Default predicate IF0. */
      bool cc : 1;
      /* Current stream out stream has overflown. */
      bool so : 1;
      /* Any stream out stream has overflown. */
      bool so_global : 1;
   } pred;

   /* The destination register is always the last element, if the instruction
    * has a destination.
    */
   struct pvr_pds_instr_src srcs[PVR_PDS_MAX_SRC_COUNT];
};

#endif /* PVR_PDS_ISA_H */
