/*
 * Copyright © 2024 Imagination Technologies Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include "nir/nir.h"
#include "nir/nir_builder.h"
#include "rogue.h"
#include "util/macros.h"

#include <assert.h>
#include <stdbool.h>
#include <stdint.h>

/**
 * \file rogue_nir_clipcull.c
 *
 * \brief Contains the rogue_nir_clipcull NIR pass.
 */

/* Vertex outputs sent to the clipping hardware can't be used by the fragment
 * shader; this pass duplicates clip/cull data so the fragment shader can use
 * it.
 */

/* TODO NEXT:
 * - Slots don't have to be contiguous if just using single floats.
 * - Probably want to either compact these, or directly clone the clip variable,
 *   regardless of it having a location_frac overlay layout or not
 * - Probably want to only populate elements actually used by consumer shader
 *   rather than just doing all of them and letting unused vars pass/compacting
 *   sort it out.
 */

struct clipcull_state {
   nir_variable *clone_var[2][8]; /* [vert, frag][elems, max 8] */
};

static bool
clone_clipcull_stores(struct nir_builder *b, nir_instr *instr, void *data)
{
   if (instr->type != nir_instr_type_intrinsic)
      return false;

   nir_intrinsic_instr *intr = nir_instr_as_intrinsic(instr);
   if (intr->intrinsic != nir_intrinsic_store_deref)
      return false;

   nir_deref_instr *deref = nir_src_as_deref(intr->src[0]);
   if (deref->deref_type != nir_deref_type_array)
      return false;

   nir_variable *var = nir_deref_instr_get_variable(deref);

   if (var->data.location != VARYING_SLOT_CLIP_DIST0 &&
       var->data.location != VARYING_SLOT_CLIP_DIST1)
      return false;

   assert(nir_src_is_const(deref->arr.index));

   unsigned arr_index = nir_src_as_int(deref->arr.index);
   unsigned var_index = var->data.location - VARYING_SLOT_CLIP_DIST0;
   unsigned flat_index = (var_index * 4) + arr_index + var->data.location_frac;

   struct clipcull_state *state = data;

   b->cursor = nir_after_instr(instr);
   nir_store_var(b, state->clone_var[0][flat_index], intr->src[1].ssa, 1);

   return true;
}

static bool is_clipcull_load(const nir_instr *instr, UNUSED const void *cb_data)
{
   if (instr->type != nir_instr_type_intrinsic)
      return false;

   nir_intrinsic_instr *intr = nir_instr_as_intrinsic(instr);
   if (intr->intrinsic != nir_intrinsic_load_deref)
      return false;

   nir_deref_instr *deref = nir_src_as_deref(intr->src[0]);
   if (deref->deref_type != nir_deref_type_array)
      return false;

   nir_variable *var = nir_deref_instr_get_variable(deref);

   if (var->data.location != VARYING_SLOT_CLIP_DIST0 &&
       var->data.location != VARYING_SLOT_CLIP_DIST1)
      return false;

   assert(nir_src_is_const(deref->arr.index));

   return true;
}

static nir_def *swap_clipcull_load(nir_builder *b, nir_instr *instr, void *cb_data)
{
   nir_intrinsic_instr *intr = nir_instr_as_intrinsic(instr);
   nir_deref_instr *deref = nir_src_as_deref(intr->src[0]);
   nir_variable *var = nir_deref_instr_get_variable(deref);

   unsigned arr_index = nir_src_as_int(deref->arr.index);
   unsigned var_index = var->data.location - VARYING_SLOT_CLIP_DIST0;
   unsigned flat_index = (var_index * 4) + arr_index + var->data.location_frac;

   struct clipcull_state *state = cb_data;

   return nir_load_var(b, state->clone_var[1][flat_index]);
}

PUBLIC
bool rogue_nir_clipcull(nir_shader *producer, nir_shader *consumer)
{
   if (!(producer->info.stage == MESA_SHADER_VERTEX && consumer->info.stage == MESA_SHADER_FRAGMENT))
      return false;

   /* Check if clip/cull vars are actually being written to. */
   unsigned dist_comps = producer->info.clip_distance_array_size + producer->info.cull_distance_array_size;
   assert(dist_comps <= 8);
   if (!dist_comps)
      return false;

   /* Check if the consumer actually reads any of the clip/cull vars, if not we
    * don't need to do anything.
    */
   bool dists_read = ((consumer->info.inputs_read & BITFIELD64_BIT(VARYING_SLOT_CLIP_DIST0)) != 0) |
                     ((consumer->info.inputs_read & BITFIELD64_BIT(VARYING_SLOT_CLIP_DIST1)) != 0);
   if (!dists_read)
      return false;

   /* Find empty varying slot(s) */
   gl_varying_slot clone_slot = ~0U;

   for (unsigned u = 0; u < (MAX_VARYING - dist_comps + 1); ++u) {
      gl_varying_slot s = VARYING_SLOT_VAR0 + u;

      /* TODO NEXT: these don't need to be contiguous. */
      /* Check for contiguous free varying slots (if needed). */
      bool found = true;
      for (unsigned v = 0; v < dist_comps; ++v)
         found &= !(producer->info.outputs_written & BITFIELD64_BIT(s + v));

      if (found) {
         clone_slot = s;
         break;
      }
   }

   assert(clone_slot != ~0U);

   struct clipcull_state state = { 0 };

   /* Create the clone varyings. */

   for (unsigned c = 0; c < dist_comps; ++c) {
      state.clone_var[0][c] = nir_variable_create(producer, nir_var_shader_out, glsl_float_type(), NULL);
      state.clone_var[0][c]->data.location = clone_slot + c;
      state.clone_var[0][c]->data.driver_location = producer->num_outputs++;

      state.clone_var[1][c] = nir_variable_create(consumer, nir_var_shader_in, glsl_float_type(), NULL);
      state.clone_var[1][c]->data.location = clone_slot + c;
      state.clone_var[1][c]->data.driver_location = consumer->num_inputs++;
   }

   /* Go to all dist stores in producer and clone them. */
   nir_shader_instructions_pass(producer, clone_clipcull_stores, nir_metadata_block_index | nir_metadata_dominance, &state);

   /* Go to all dist loads in consumer and swap them to use clone varying. */
   nir_shader_lower_instructions(consumer, is_clipcull_load, swap_clipcull_load, &state);

   return true;
}
