/*
 * Copyright © 2024 Imagination Technologies Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include "nir/nir.h"
#include "nir/nir_builder.h"
#include "nir/nir_builtin_builder.h"
#include "nir/nir_search_helpers.h"
#include "rogue.h"
#include "util/macros.h"

/**
 * \file rogue_nir_lower_xfb.c
 *
 * \brief Contains transform feedback lowering pass.
 */

struct xfb_state {
   nir_def *xfb_state_base_addr;
   nir_def *is_active;

   struct {
      nir_def *base_addr;
      nir_def *base_offset;
      nir_def *size;
      nir_def *non_null;
      nir_def *bytes_written_addr;
   } xfb_buffer[PVR_MAX_XFB_BUFFERS];
};

/* Helpers. */
static inline nir_def *nir_load64(nir_builder *b, nir_def *addr)
{
   return nir_load_global(b, addr, 8, 1, 64);
}

static inline nir_def *nir_load32(nir_builder *b, nir_def *addr)
{
   return nir_load_global(b, addr, 4, 1, 32);
}

static inline void nir_store32(nir_builder *b, nir_def *addr, nir_def *value)
{
   assert(value->bit_size == 32);
   assert(value->num_components == 1);

   nir_store_global(b, addr, 4, value, 1);
}

static nir_def *nir_atomic_add32(nir_builder *b, nir_def *addr, nir_def *value)
{
   return nir_global_atomic(b, 32, addr, value, .atomic_op = nir_atomic_op_iadd);
}

static nir_def *nir_atomic_add32_imm(nir_builder *b, nir_def *addr, uint32_t value)
{
   return nir_global_atomic(b, 32, addr, nir_imm_int(b, value), .atomic_op = nir_atomic_op_iadd);
}

static nir_def *nir_atomic_ior32_imm(nir_builder *b, nir_def *addr, uint32_t value)
{
   return nir_global_atomic(b, 32, addr, nir_imm_int(b, value), .atomic_op = nir_atomic_op_ior);
}

#define rogue_xfb_state_offset(member) \
   nir_iadd_imm(b, state->xfb_state_base_addr, offsetof(rogue_xfb_state, member))

#define rogue_xfb_state_info_field(field)                           \
   nir_ubitfield_extract_imm(b,                                     \
                             info_word,                             \
                             rogue_xfb_state_info_##field##_offset, \
                             rogue_xfb_state_info_##field##_size)

static inline nir_def *buffer_non_null(nir_def *info_word, nir_builder *b, unsigned buffer)
{
   assert(buffer < rogue_xfb_state_info_buffer_mask_size);

   return
      nir_i2b(b,
         nir_ubitfield_extract_imm(b,
                                   info_word,
                                   rogue_xfb_state_info_buffer_mask_offset + buffer,
                                   1));
}

static inline nir_def *counter_buffer_non_null(nir_def *info_word, nir_builder *b, unsigned buffer)
{
   assert(buffer < rogue_xfb_state_info_counter_buffer_mask_size);

   return
      nir_i2b(b,
         nir_ubitfield_extract_imm(b,
                                   info_word,
                                   rogue_xfb_state_info_counter_buffer_mask_offset + buffer,
                                   1));
}

static inline nir_def *last_bytes_written_addr_non_null(nir_def *info_word, nir_builder *b, unsigned buffer)
{
   assert(buffer < rogue_xfb_state_info_last_bytes_written_addr_mask_size);

   return
      nir_i2b(b,
         nir_ubitfield_extract_imm(b,
                                   info_word,
                                   rogue_xfb_state_info_last_bytes_written_addr_mask_offset + buffer,
                                   1));
}

static void setup_xfb_state(struct xfb_state *state, nir_builder *b)
{
   state->xfb_state_base_addr = nir_load_xfb_address(b, 64);

   nir_def *info_word = nir_load32(b, rogue_xfb_state_offset(info));
   state->is_active = nir_i2b(b, rogue_xfb_state_info_field(active));

   nir_if *nif_is_active = nir_push_if(b, state->is_active);
   {
      nir_def *vert_id = nir_isub(b, nir_load_vertex_id(b), nir_load_base_vertex(b));
      nir_def *inst_id = nir_isub(b, nir_load_instance_id(b), nir_load_base_instance(b));

      /* nir_def *num_verts = nir_load_num_vertices(b); */ /* TODO */
      nir_def *num_verts = nir_load32(b, rogue_xfb_state_offset(num_vertices));

      nir_def *per_inst_offset = nir_imul(b, inst_id, num_verts);
      nir_def *per_vert_offset = nir_iadd(b, per_inst_offset, vert_id);

      nir_def *imm_zero = nir_imm_int(b, 0);
      nir_def *is_first_inst = nir_ieq(b, per_vert_offset, imm_zero);

      u_foreach_bit (buffer, b->shader->xfb_info->buffers_written) {
         unsigned stride = b->shader->info.xfb_stride[buffer] * 4;
         state->xfb_buffer[buffer].base_addr = nir_load64(b, rogue_xfb_state_offset(buffer[buffer].addr));
         state->xfb_buffer[buffer].size = nir_load32(b, rogue_xfb_state_offset(buffer[buffer].size));
         state->xfb_buffer[buffer].non_null = buffer_non_null(info_word, b, buffer);
         state->xfb_buffer[buffer].bytes_written_addr = rogue_xfb_state_offset(buffer[buffer].bytes_written);

         nir_def *uncounted_base_offset;
         nir_def *counted_base_offset;
         nir_if *nif_counter_buffer_non_null = nir_push_if(b, counter_buffer_non_null(info_word, b, buffer));
         {
            nir_def *base_offset = nir_imul_imm(b, per_vert_offset, stride);
            nir_def *counter_buffer_addr = nir_load64(b, rogue_xfb_state_offset(buffer[buffer].counter_buffer_addr));
            nir_def *counter_offset = nir_load32(b, counter_buffer_addr);
            counted_base_offset = nir_iadd(b, base_offset, counter_offset);
         }
         nir_push_else(b, nif_counter_buffer_non_null);
         {
            nir_def *base_offset = nir_imul_imm(b, per_vert_offset, stride);
            uncounted_base_offset = base_offset;
         }
         nir_pop_if(b, nif_counter_buffer_non_null);

         nir_phi_instr *phi = nir_phi_instr_create(b->shader);
         nir_phi_instr_add_src(phi, nir_if_last_then_block(nif_counter_buffer_non_null), counted_base_offset);
         nir_phi_instr_add_src(phi, nir_if_last_else_block(nif_counter_buffer_non_null), uncounted_base_offset);
         nir_def_init(&phi->instr, &phi->def, 1, 32);
         nir_builder_instr_insert(b, &phi->instr);

         state->xfb_buffer[buffer].base_offset = &phi->def;

         /* Accumulate bytes written. */
         nir_if *nif_1st_and_lbwa_non_null = nir_push_if(b, nir_iand(b, is_first_inst, last_bytes_written_addr_non_null(info_word, b, buffer)));
         {
            nir_def *last_bytes_written_addr = nir_load64(b, rogue_xfb_state_offset(buffer[buffer].last_bytes_written_addr));
            nir_def *last_bytes_written = nir_load32(b, last_bytes_written_addr);
            nir_atomic_add32(b, state->xfb_buffer[buffer].bytes_written_addr, last_bytes_written);
         }
         nir_pop_if(b, nif_1st_and_lbwa_non_null);
      }

      nir_def *query_pool_active = nir_i2b(b, rogue_xfb_state_info_field(query_pool_active));
      nir_if *nif_query_pool_active = nir_push_if(b, query_pool_active);
      {
         nir_def *query_pool_result_addr = nir_load64(b, rogue_xfb_state_offset(query_pool_result_addr));

         nir_def *verts_per_prim = rogue_xfb_state_info_field(verts_per_prim);
         nir_def *increment = nir_inot(b, nir_i2b(b, nir_umod(b, vert_id, verts_per_prim)));
         nir_if *nif_increment = nir_push_if(b, increment);
         {
            /* TODO: this only increments the first 64-bit value, and even then,
             * only the lower 32-bits.
             */
            nir_atomic_add32_imm(b, nir_iadd_imm(b, query_pool_result_addr, 0), 1);
            /* nir_atomic_add32_imm(b, nir_iadd_imm(b, query_pool_result_addr, 8), 1); */
         }
         nir_pop_if(b, nif_increment);
      }
      nir_pop_if(b, nif_query_pool_active);
   }
   nir_pop_if(b, nif_is_active);
}

static bool
lower_xfb(struct nir_builder *b, nir_intrinsic_instr *intr, void *data)
{
   struct xfb_state *state = (struct xfb_state *)data;

   if (intr->intrinsic != nir_intrinsic_store_output)
      return false;

   assert(nir_intrinsic_write_mask(intr) == 1);

   nir_def *value = intr->src[0].ssa;
   assert(value->bit_size == 32);
   assert(value->num_components == 1);
   assert(!nir_src_as_uint(intr->src[1]));
   const unsigned value_size = value->bit_size / 8;

   unsigned comp = nir_intrinsic_component(intr);
   nir_io_xfb xfb = comp < 2 ? nir_intrinsic_io_xfb(intr) : nir_intrinsic_io_xfb2(intr);
   if (!xfb.out[comp % 2].num_components)
      return false;

   assert(xfb.out[comp % 2].num_components == 1);

   unsigned buffer = xfb.out[comp % 2].buffer;
   assert(buffer < PVR_MAX_XFB_BUFFERS);

   unsigned offset = xfb.out[comp % 2].offset * 4;

   b->cursor = nir_before_instr(&intr->instr);

   nir_if *nif_is_active = nir_push_if(b, state->is_active);
   {
      nir_def *buffer_offset = nir_iadd_imm(b, state->xfb_buffer[buffer].base_offset, offset);
      nir_def *buffer_offset_in_range = nir_ilt(b, nir_iadd_imm(b, buffer_offset, value_size - 1), state->xfb_buffer[buffer].size);

      nir_def *xfb_write_cond = nir_iand(b, state->xfb_buffer[buffer].non_null, buffer_offset_in_range);
      nir_if *nif_xfb_write = nir_push_if(b, xfb_write_cond);
      {
         nir_def *write_addr = nir_iadd(b, state->xfb_buffer[buffer].base_addr, nir_u2u64(b, buffer_offset));
         nir_store32(b, write_addr, value);
         nir_atomic_add32_imm(b, state->xfb_buffer[buffer].bytes_written_addr, value_size);
      }
      nir_pop_if(b, nif_xfb_write);
   }
   nir_pop_if(b, nif_is_active);

   /* nir_instr_remove(&intr->instr); */

   return true;
}

PUBLIC
bool rogue_nir_lower_xfb(nir_shader *shader, rogue_build_ctx *ctx)
{
   assert(shader->info.stage == MESA_SHADER_VERTEX);

   struct xfb_state state = { 0 };

   nir_function_impl *entry = nir_shader_get_entrypoint(shader);
   nir_builder b = nir_builder_create(entry);
   b.cursor = nir_before_block(nir_start_block(entry));

   setup_xfb_state(&state, &b);

   nir_shader_intrinsics_pass(shader, lower_xfb, nir_metadata_none, &state);

   return true;
}

