/*
 * Copyright © 2024 Imagination Technologies Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef PVR_PDS_BUILDER_H
#define PVR_PDS_BUILDER_H

#include <stdint.h>

#include "pvr_pds_ra.h"
#include "pvr_types.h"
#include "util/macros.h"

#define __pvr_address_type pvr_dev_addr_t
#define __pvr_get_address(pvr_dev_addr) (pvr_dev_addr).addr
/* clang-format off */
#define __pvr_make_address(addr_u64) PVR_DEV_ADDR(addr_u64)
/* clang-format on */

#include "pds/rogue_pds_instr_srcs.h"

/* TODO: dedup these with csbgen? */
#define pvr_instr_src(x) CONCAT2(ROGUE_PDS_INSTR_SRC_, x)
#define pvr_instr_src_length(x) CONCAT2(pvr_instr_src(x), _length)
#define pvr_instr_src_header(x) CONCAT2(pvr_instr_src(x), _header)
#define pvr_instr_src_pack_raw(x) CONCAT2(pvr_instr_src(x), _pack)

#define pvr_instr_src_pack(_dst, cmd, name)                             \
   for (struct pvr_instr_src(cmd) name = { pvr_instr_src_header(cmd) }, \
                                  *_loop_terminate = &name;             \
        __builtin_expect(_loop_terminate != NULL, 1);                   \
        ({                                                              \
           STATIC_ASSERT(sizeof(*(_dst)) ==                             \
                         PVR_DW_TO_BYTES(pvr_instr_src_length(cmd)));   \
           pvr_instr_src_pack_raw(cmd)((_dst), &name);                  \
           _loop_terminate = NULL;                                      \
        }))

#define pvr_pds_allocate_src_const(_instr_src, _ra_state) \
   ((pvr_instr_src_length(_instr_src) == 1)               \
       ? pvr_pds_ra_allocate_32b_const((_ra_state))       \
       : pvr_pds_ra_allocate_64b_const((_ra_state)))

#define pvr_pds_allocate_src_const_contiguous(_instr_src, _ra_state, _count) \
   ((pvr_instr_src_length(_instr_src) == 1)                                  \
       ? pvr_pds_ra_allocate_32b_const_contiguous((_ra_state), (_count))     \
       : pvr_pds_ra_allocate_64b_const_contiguous((_ra_state), (_count)))

#define pvr_pds_allocate_src_temp(_instr_src, _ra_state) \
   ((pvr_instr_src_length(_instr_src) == 1)               \
       ? pvr_pds_ra_allocate_32b_temp((_ra_state))       \
       : pvr_pds_ra_allocate_64b_temp((_ra_state)))

#define pvr_pds_allocate_src_temp_contiguous(_instr_src, _ra_state, _count) \
   ((pvr_instr_src_length(_instr_src) == 1)                                  \
       ? pvr_pds_ra_allocate_32b_temp_contiguous((_ra_state), (_count))     \
       : pvr_pds_ra_allocate_64b_temp_contiguous((_ra_state), (_count)))

struct pvr_device_info;
struct pvr_pds_instr;

void pvr_pds_build_mad(struct pvr_pds_instr *instr,
                       const struct pvr_pds_reg dst,
                       const struct pvr_pds_reg mul0,
                       const struct pvr_pds_reg mul1,
                       const struct pvr_pds_reg add,
                       uint32_t flags);

void pvr_pds_build_sftlp32(struct pvr_pds_instr *instr,
                           enum pvr_pds_logic_op_type logic_op,
                           const struct pvr_pds_reg dst,
                           const struct pvr_pds_reg src0,
                           const struct pvr_pds_reg src1,
                           const struct pvr_pds_reg shift);
void pvr_pds_build_sftlp32i(struct pvr_pds_instr *instr,
                            enum pvr_pds_logic_op_type logic_op,
                            const struct pvr_pds_reg dst,
                            const struct pvr_pds_reg src0,
                            const struct pvr_pds_reg src1,
                            int32_t shift);
void pvr_pds_build_sftlp64(struct pvr_pds_instr *instr,
                           enum pvr_pds_logic_op_type logic_op,
                           const struct pvr_pds_reg dst,
                           const struct pvr_pds_reg src0,
                           const struct pvr_pds_reg src1,
                           const struct pvr_pds_reg shift);
void pvr_pds_build_sftlp64i(struct pvr_pds_instr *instr,
                            enum pvr_pds_logic_op_type logic_op,
                            const struct pvr_pds_reg dst,
                            const struct pvr_pds_reg src0,
                            const struct pvr_pds_reg src1,
                            int32_t shift);

/* Logic-OPeration 32. */
static inline void pvr_pds_build_lop32(struct pvr_pds_instr *instr,
                                       enum pvr_pds_logic_op_type logic_op,
                                       const struct pvr_pds_reg dst,
                                       const struct pvr_pds_reg src0,
                                       const struct pvr_pds_reg src1)
{
   pvr_pds_build_sftlp32i(instr, logic_op, dst, src0, src1, 0);
}
/* Logic-OPeration 64. */
static inline void pvr_pds_build_lop64(struct pvr_pds_instr *instr,
                                       enum pvr_pds_logic_op_type logic_op,
                                       const struct pvr_pds_reg dst,
                                       const struct pvr_pds_reg src0,
                                       const struct pvr_pds_reg src1)
{
   pvr_pds_build_sftlp64i(instr, logic_op, dst, src0, src1, 0);
}

/* Copy a 32b temp register's contents to another 32b temp. */
static void pvr_pds_build_mov32(struct pvr_pds_instr *instr,
                                const struct pvr_pds_reg dst,
                                const struct pvr_pds_reg src)
{
   pvr_pds_build_lop32(instr, PVR_PDS_LOGIC_OP_NONE, dst, src, src);
}
/* Copy a 64 P/Temp register's contents to another 64b P/Temp. */
static void pvr_pds_build_mov64(struct pvr_pds_instr *instr,
                                const struct pvr_pds_reg dst,
                                const struct pvr_pds_reg src)
{
   pvr_pds_build_lop64(instr, PVR_PDS_LOGIC_OP_NONE, dst, src, src);
}

void pvr_pds_build_stm(struct pvr_pds_instr *instr,
                       uint32_t so_idx,
                       const struct pvr_pds_reg src0,
                       const struct pvr_pds_reg src1,
                       const struct pvr_pds_reg src2,
                       const struct pvr_pds_reg src3,
                       uint32_t flags);
void pvr_pds_build_stmc(struct pvr_pds_instr *instr, uint32_t so_mask);

void pvr_pds_build_add32(ASSERTED const struct pvr_device_info *dev_info,
                         struct pvr_pds_instr *instr,
                         const struct pvr_pds_reg dst,
                         const struct pvr_pds_reg src0,
                         const struct pvr_pds_reg src1,
                         uint32_t flags);
void pvr_pds_build_add64(struct pvr_pds_instr *instr,
                         const struct pvr_pds_reg dst,
                         const struct pvr_pds_reg src0,
                         const struct pvr_pds_reg src1,
                         uint32_t flags);

void pvr_pds_build_cmp(struct pvr_pds_instr *instr,
                       enum pvr_pds_comparison_op_type compare_op,
                       const struct pvr_pds_reg src0,
                       const struct pvr_pds_reg src1);

void pvr_pds_build_bra(struct pvr_pds_instr *instr,
                       enum pvr_pds_predicate src_pred,
                       enum pvr_pds_predicate set_pred,
                       uint32_t branch_offset,
                       uint32_t flags);

void pvr_pds_build_ddmad(struct pvr_pds_instr *instr,
                         const struct pvr_pds_reg src0,
                         const struct pvr_pds_reg src1,
                         const struct pvr_pds_reg src2,
                         const struct pvr_pds_reg src3,
                         uint32_t flags);

void pvr_pds_build_doutu(struct pvr_pds_instr *instr,
                         const struct pvr_pds_reg src0,
                         uint32_t flags);
void pvr_pds_build_doutd(struct pvr_pds_instr *instr,
                         const struct pvr_pds_reg src0,
                         const struct pvr_pds_reg src1,
                         uint32_t flags);
void pvr_pds_build_doutw(struct pvr_pds_instr *instr,
                         const struct pvr_pds_reg src0,
                         const struct pvr_pds_reg src1,
                         uint32_t flags);
void pvr_pds_build_doutv(struct pvr_pds_instr *instr,
                         const struct pvr_pds_reg src0,
                         const struct pvr_pds_reg src1,
                         uint32_t flags);
void pvr_pds_build_doutc(struct pvr_pds_instr *instr,
                         const struct pvr_pds_reg src0,
                         const struct pvr_pds_reg src1,
                         uint32_t flags);
void pvr_pds_build_douti(struct pvr_pds_instr *instr,
                         const struct pvr_pds_reg src0,
                         const struct pvr_pds_reg src1,
                         uint32_t flags);

void pvr_pds_build_limm(struct pvr_pds_instr *instr,
                        const struct pvr_pds_reg dst,
                        uint16_t imm16);

void pvr_pds_build_limm64(struct pvr_pds_instr *instr,
                          const struct pvr_pds_reg dst,
                          uint64_t imm64);

struct pvr_pds_reg
pvr_pds_build_convert_32_to_64(const struct pvr_pds_reg reg_low,
                               const struct pvr_pds_reg reg_high);

/* Load Global Register. */
void pvr_pds_build_lgr(ASSERTED const struct pvr_device_info *dev_info,
                       struct pvr_pds_instr *instr,
                       const struct pvr_pds_reg dst,
                       uint32_t src_global_reg_idx);

void pvr_pds_build_ld(struct pvr_pds_instr *instr,
                      const struct pvr_pds_reg src0);
void pvr_pds_build_st(struct pvr_pds_instr *instr,
                      const struct pvr_pds_reg src0);

void pvr_pds_build_wdf(struct pvr_pds_instr *instr);
void pvr_pds_build_lock(struct pvr_pds_instr *instr);
void pvr_pds_build_release(struct pvr_pds_instr *instr);
void pvr_pds_build_halt(struct pvr_pds_instr *instr);
void pvr_pds_build_nop(struct pvr_pds_instr *instr);

static void pvr_pds_src_write_32b_const(void *buffer,
                                        const struct pvr_pds_reg_array *reg_arr,
                                        const void *src,
                                        uint32_t count)
{
   uint32_t *buffer_u32 = (uint32_t *)buffer;

   assert(reg_arr->reg0.reg_size == PVR_PDS_REG_SIZE_32);
   assert(reg_arr->reg0.idx < 128);
   assert(count <= reg_arr->reg_count);

   memcpy(&buffer_u32[reg_arr->reg0.idx], src, sizeof(*buffer_u32) * count);
}

static void pvr_pds_src_write_64b_const(void *buffer,
                                        const struct pvr_pds_reg_array *reg_arr,
                                        const void *src,
                                        uint32_t count)
{
   uint64_t *buffer_u64 = (uint64_t *)buffer;

   assert(reg_arr->reg0.reg_size == PVR_PDS_REG_SIZE_64);
   assert(reg_arr->reg0.idx < 128 / 2);
   assert(count <= reg_arr->reg_count);

   memcpy(&buffer_u64[reg_arr->reg0.idx], src, sizeof(*buffer_u64) * count);
}

#define pvr_pds_src_write_multiple(_instr_src, _buffer, _reg_arr, _src, _count) \
   do {                                                                         \
      const struct pvr_pds_reg_array *reg_arr = (_reg_arr);                     \
      assert(reg_arr->reg0.type == PVR_PDS_REG_TYPE_CONST);                     \
      STATIC_ASSERT(sizeof(*(_src)) ==                                          \
                    PVR_DW_TO_BYTES(pvr_instr_src_length(_instr_src)));         \
      _Generic((_src), \
               uint32_t *: \
                  pvr_pds_src_write_32b_const((_buffer), reg_arr, (_src), (_count)),  \
               const uint32_t *: \
                  pvr_pds_src_write_32b_const((_buffer), reg_arr, (_src), (_count)),  \
               uint64_t *:  \
                  pvr_pds_src_write_64b_const((_buffer), reg_arr, (_src), (_count)),  \
               const uint64_t *:  \
                  pvr_pds_src_write_64b_const((_buffer), reg_arr, (_src), (_count)));                                                        \
   } while (0)

#define pvr_pds_src_write(_instr_src, _buffer, _reg, _src)                      \
   do {                                                                         \
      struct pvr_pds_reg_array reg_array = { .reg0 = (_reg), .reg_count = 1 };  \
      pvr_pds_src_write_multiple(_instr_src, (_buffer), &reg_array, (_src), 1); \
   } while (0)

static enum pvr_instr_src(DOUTW_BSIZE)
   pvr_pds_doutw_src1_bsize_from_uint(uint32_t size)
{
   switch (size) {
   case 0:
      return pvr_instr_src(DOUTW_BSIZE_NONE);
   case 1:
      return pvr_instr_src(DOUTW_BSIZE_LOWER);
   case 2:
      return pvr_instr_src(DOUTW_BSIZE_ALL64);
   default:
      unreachable("Invalid size for doutw");
   }
}

#endif /* PVR_PDS_BUILDER_H */
