/*
 * Copyright © 2024 Imagination Technologies Ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <stdbool.h>
#include <stdint.h>
#include <vulkan/vulkan.h>

#include "hwdef/rogue_hw_utils.h"
#include "pvr_private.h"
#include "pvr_types.h"
#include "pvr_pds_gen.h"
#include "vk_command_buffer.h"
#include "vk_alloc.h"
#include "util/u_math.h"

static void pvr_create_xfb_streamout_program_config(
   struct pvr_cmd_buffer *cmd_buffer,
   struct pvr_pds_xfb_program_input *input_config)
{
   struct pvr_cmd_buffer_xfb_state *recording_state =
      &cmd_buffer->state.xfb_state;
   struct pvr_graphics_pipeline *pipeline = cmd_buffer->state.gfx_pipeline;
   struct pvr_xfb_shader_state *xfb_shader_state =
      &pipeline->shader_state.xfb_shader_state;
   const VkAllocationCallbacks *const allocator = &cmd_buffer->device->vk.alloc;
   struct pvr_xfb_binding *xfb_bindings;
   uint32_t varying_counters = 0;

   input_config->dev_info = &cmd_buffer->device->pdevice->dev_info;
   input_config->test_overflow = true;
   input_config->primitive_type =
      cmd_buffer->vk.dynamic_graphics_state.ia.primitive_topology;
   input_config->varying_count = xfb_shader_state->total_varying_count;

   input_config->data.buffer_info =
      vk_alloc(allocator,
               sizeof(struct pvr_pds_xfb_buffer_info) *
                  xfb_shader_state->total_varying_count,
               8,
               VK_SYSTEM_ALLOCATION_SCOPE_DEVICE);

   if (!input_config->data.buffer_info) {
      vk_error(allocator, VK_ERROR_OUT_OF_HOST_MEMORY);
      return;
   }

   input_config->data.varying_info_arr =
      vk_alloc(allocator,
               sizeof(struct pvr_pds_xfb_varying_info) *
                  xfb_shader_state->total_varying_count,
               8,
               VK_SYSTEM_ALLOCATION_SCOPE_DEVICE);

   if (!input_config->data.varying_info_arr) {
      vk_error(allocator, VK_ERROR_OUT_OF_HOST_MEMORY);
      goto varying_info_arr_fail;
   }

   input_config->data.counter_buffer_entries =
      vk_alloc(allocator,
               sizeof(struct pvr_pds_xfb_counter_buffer_info) *
                  xfb_shader_state->total_varying_count,
               8,
               VK_SYSTEM_ALLOCATION_SCOPE_DEVICE);

   if (!input_config->data.counter_buffer_entries) {
      vk_error(allocator, VK_ERROR_OUT_OF_HOST_MEMORY);
      goto counter_buffer_entries_fail;
   }

   for (uint32_t uiStream = 0; uiStream < PVR_MAX_XFB_STREAMS; uiStream++) {
      struct pvr_xfb_varying *varyings =
         xfb_shader_state->streams[uiStream].varyings;
      uint32_t varying_count =
         xfb_shader_state->streams[uiStream].varying_count;

      input_config->varying_number[uiStream] = varying_count;

      if (!varying_count) {
         continue;
      }

      for (uint32_t i = 0; i < varying_count; i++) {
         uint32_t vindex = varying_counters + i;
         uint32_t vbuff_index = varyings[i].xfb_buffer;
         input_config->data;

         // Varying info
         {
            input_config->data.varying_info_arr[vindex].input_offset =
               varyings[i].register_start;
            input_config->data.varying_info_arr[vindex].output_offset =
               varyings[i].xfb_offset;
            input_config->data.varying_info_arr[vindex].varying_size =
               varyings[i].xfb_stride;
         }

         // Gather counter buffer info
         input_config->data.counter_buffer_entries[vindex]
            .counter_buffer_varying_index = vbuff_index;

         // Buffer info
         xfb_bindings = &recording_state->bindings[vbuff_index];
         if (xfb_bindings) {
            // Offset and limit adujust at xfb buffer bind API
            input_config->data.buffer_info[vindex].buffer_address =
               xfb_bindings->buffer_addr;
            input_config->data.buffer_info[vindex].buffer_size =
               xfb_bindings->buffer_limit;
         }
      }
      varying_counters += varying_count;
   }

   input_config->persistent_temp_size = input_config->varying_count * 3;
   return;
counter_buffer_entries_fail:
   vk_free(allocator, input_config->data.varying_info_arr);
varying_info_arr_fail:
   vk_free(allocator, input_config->data.buffer_info);
   return;
}

static void
pvr_setup_init_config(struct pvr_cmd_buffer *cmd_buffer,
                      struct pvr_pds_xfb_init_program_input *init_input,
                      struct pvr_pds_xfb_program_input *input_config)
{
   struct pvr_cmd_buffer_xfb_state *psXFBState = &cmd_buffer->state.xfb_state;
   const VkAllocationCallbacks *const allocator = &cmd_buffer->device->vk.alloc;

   if (psXFBState->counter_buffer_count) {
      init_input->load_from_counter_buffers = true;
      init_input->counter_buffer_count = psXFBState->counter_buffer_count;
      init_input->data.counter_buffers_addr =
         vk_alloc(allocator,
                  sizeof(pvr_dev_addr_t) * init_input->counter_buffer_count,
                  8U,
                  VK_SYSTEM_ALLOCATION_SCOPE_COMMAND);

      for (uint32_t i = 0; i < init_input->counter_buffer_count; i++) {
         if (!psXFBState->bindings[i].counter_buffer_addr.addr) {
            init_input->data.counter_buffers_addr[i].addr = 0;
         } else {
            init_input->data.counter_buffers_addr[i].addr =
               psXFBState->bindings[i].counter_buffer_addr.addr;
         }
      }

      init_input->num_loads = input_config->varying_count;
      init_input->load_dest_select =
         vk_alloc(allocator,
                  sizeof(uint32_t) * init_input->num_loads,
                  8U,
                  VK_SYSTEM_ALLOCATION_SCOPE_COMMAND);

      for (uint32_t i = 0; i < init_input->num_loads; i++) {
         init_input->load_dest_select[i] =
            input_config->data.counter_buffer_entries[i]
               .counter_buffer_varying_index;
      }
   } else {
      init_input->load_from_counter_buffers = false;
   }

   if (input_config->data.buffer_info)
      init_input->data.buffer_info = input_config->data.buffer_info;

   if (input_config->data.varying_info_arr)
      init_input->data.varying_info_arr = input_config->data.varying_info_arr;

   init_input->dev_info = input_config->dev_info;
   init_input->data.counter_pairs_address =
      &cmd_buffer->state.gfx_pipeline->shader_state.xfb_shader_state.counters
          ->dev_addr;

   init_input->data.counter_buffer_offsets_addr =
      &cmd_buffer->state.gfx_pipeline->shader_state.xfb_shader_state
          .counter_buffer_offsets->dev_addr;

   init_input->primitive_type = input_config->primitive_type;

   init_input->persistent_temp_size = input_config->persistent_temp_size;
   init_input->data.ptemp_memory_buffer_addr =
      input_config->data.ptemp_memory_buffer_addr;
}

static void
pvr_setup_end_program_config(struct pvr_cmd_buffer *cmd_buffer,
                             struct pvr_pds_xfb_end_program_input *input)
{
   struct pvr_cmd_buffer_xfb_state *psXFBState = &cmd_buffer->state.xfb_state;
   const VkAllocationCallbacks *const allocator = &cmd_buffer->device->vk.alloc;

   struct pvr_graphics_pipeline *pipeline;
   pipeline = cmd_buffer->state.gfx_pipeline;
   struct pvr_xfb_shader_state *xfb_shader_state;
   xfb_shader_state = &pipeline->shader_state.xfb_shader_state;

   if (psXFBState->counter_buffer_count) {
      input->store_counter_buffers = true;
      input->counter_buffer_count = psXFBState->counter_buffer_count;

      input->data.counter_buffers_addr =
         vk_alloc(allocator,
                  sizeof(pvr_dev_addr_t) * input->counter_buffer_count,
                  8U,
                  VK_SYSTEM_ALLOCATION_SCOPE_COMMAND);

      for (uint32_t i = 0; i < input->counter_buffer_count; i++) {
         pvr_dev_addr_t *counter_buffers_addr =
            &input->data.counter_buffers_addr[i];

         if (!psXFBState->bindings[i].counter_buffer_addr.addr) {
            counter_buffers_addr->addr = 0;
         } else {
            counter_buffers_addr->addr =
               psXFBState->bindings[i].counter_buffer_addr.addr;

            input->buffer_strides[i] =
               xfb_shader_state->streams[0].buffer_strides[i];

            counter_buffers_addr->addr =
               psXFBState->bindings[i].counter_buffer_addr.addr;
         }
      }
   } else {
      input->store_counter_buffers = false;
   }

   input->data.counter_pairs_address = &xfb_shader_state->counters->dev_addr;

   input->data.counter_buffer_offsets_addr =
      &xfb_shader_state->counter_buffer_offsets->dev_addr;
}

void pvr_CmdBindTransformFeedbackBuffersEXT(VkCommandBuffer commandBuffer,
                                            uint32_t firstBinding,
                                            uint32_t bindingCount,
                                            const VkBuffer *pBuffers,
                                            const VkDeviceSize *pOffsets,
                                            const VkDeviceSize *pSizes)
{
   PVR_FROM_HANDLE(pvr_cmd_buffer, cmd_buffer, commandBuffer);
   struct pvr_cmd_buffer_xfb_state *xfb_state = &cmd_buffer->state.xfb_state;

   PVR_CHECK_COMMAND_BUFFER_BUILDING_STATE(cmd_buffer);

   for (uint32_t i = 0; i < PVR_MAX_XFB_BUFFERS; i++) {
      struct pvr_xfb_binding *binding = &xfb_state->bindings[i];

      if (i < firstBinding || i >= bindingCount) {
         memset(&binding->buffer_addr, 0, sizeof(binding->buffer_addr));
         memset(&binding->buffer_limit, 0, sizeof(binding->buffer_limit));
         binding->buffer_size = 0;
         continue;
      }

      unsigned j = i - firstBinding;
      PVR_FROM_HANDLE(pvr_buffer, buffer, pBuffers[j]);

      binding->buffer_addr = PVR_DEV_ADDR_OFFSET(buffer->dev_addr, pOffsets[j]);

      if (!pSizes || pSizes[j] == VK_WHOLE_SIZE) {
         binding->buffer_limit =
            PVR_DEV_ADDR_OFFSET(buffer->dev_addr, buffer->vk.size);
         binding->buffer_size = buffer->vk.size;
      } else {
         binding->buffer_limit =
            PVR_DEV_ADDR_OFFSET(binding->buffer_addr, pSizes[j]);
         binding->buffer_size = pSizes[j];
      }
   }
}

void pvr_CmdBeginTransformFeedbackEXT(VkCommandBuffer commandBuffer,
                                      uint32_t firstCounterBuffer,
                                      uint32_t counterBufferCount,
                                      const VkBuffer *pCounterBuffers,
                                      const VkDeviceSize *pCounterBufferOffsets)
{
   PVR_FROM_HANDLE(pvr_cmd_buffer, cmd_buffer, commandBuffer);
   struct pvr_cmd_buffer_xfb_state *xfb_state = &cmd_buffer->state.xfb_state;

   PVR_CHECK_COMMAND_BUFFER_BUILDING_STATE(cmd_buffer);
   xfb_state->counter_buffer_count = counterBufferCount;

   for (uint32_t i = 0; i < PVR_MAX_XFB_BUFFERS; i++) {
      struct pvr_xfb_binding *binding = &xfb_state->bindings[i];

      if (i < firstCounterBuffer || i >= counterBufferCount) {
         memset(&binding->counter_buffer_addr, 0, sizeof(binding->counter_buffer_addr));
         continue;
      }

      unsigned j = i - firstCounterBuffer;
      PVR_FROM_HANDLE(pvr_buffer, buffer, pCounterBuffers[j]);
      pvr_dev_addr_t addr;

      if (buffer) {
         addr = buffer->dev_addr;

         if (pCounterBufferOffsets)
            addr = PVR_DEV_ADDR_OFFSET(addr, pCounterBufferOffsets[j]);

         binding->counter_buffer_addr = addr;
      } else {
         memset(&binding->counter_buffer_addr,
                0,
                sizeof(binding->counter_buffer_addr));
      }
   }

   xfb_state->is_active = true;

#if 0
   if (cmd_buffer->state.gfx_pipeline->shader_state.xfb_shader_state
          .total_varying_count) {
      pvr_kick_off_xfb(cmd_buffer);
   }
#endif
}

void pvr_CmdEndTransformFeedbackEXT(VkCommandBuffer commandBuffer,
                                    uint32_t firstCounterBuffer,
                                    uint32_t counterBufferCount,
                                    const VkBuffer *pCounterBuffers,
                                    const VkDeviceSize *pCounterBufferOffsets)
{
   PVR_FROM_HANDLE(pvr_cmd_buffer, cmd_buffer, commandBuffer);
   enum pvr_sub_cmd_type last_sub_cmd_type = PVR_SUB_CMD_TYPE_INVALID;
   struct pvr_cmd_buffer_xfb_state *xfb_state;
   ASSERTED VkResult result;

   xfb_state = &cmd_buffer->state.xfb_state;
   xfb_state->counter_buffer_count = counterBufferCount;

   struct pvr_query_info query_info = {
      .type = PVR_QUERY_TYPE_XFB_END,
   };

   for (uint32_t i = 0; i < PVR_MAX_XFB_BUFFERS; i++) {
      struct pvr_xfb_binding *binding = &xfb_state->bindings[i];

      if (i < firstCounterBuffer || i >= counterBufferCount) {
         memset(&binding->last_counter_buffer_addr, 0, sizeof(binding->last_counter_buffer_addr));
         continue;
      }

      unsigned j = i - firstCounterBuffer;
      PVR_FROM_HANDLE(pvr_buffer, buffer, pCounterBuffers[j]);
      pvr_dev_addr_t addr;

      if (buffer) {
         addr = buffer->dev_addr;

         if (pCounterBufferOffsets)
            addr = PVR_DEV_ADDR_OFFSET(addr, pCounterBufferOffsets[j]);

         binding->last_counter_buffer_addr = addr;

         query_info.xfb_end.counter_buffer_mask |= BITFIELD_BIT(i);
         query_info.xfb_end.last_bytes_written_addrs[i] = xfb_state->last_bytes_written_addrs[i].addr;
         query_info.xfb_end.end_counter_buffer_addrs[i] = addr.addr;
      } else {
         memset(&binding->last_counter_buffer_addr,
                0,
                sizeof(binding->last_counter_buffer_addr));
      }
   }
   xfb_state->is_active = false;

   if (cmd_buffer->state.current_sub_cmd) {
      if (cmd_buffer->state.current_sub_cmd->type == PVR_SUB_CMD_TYPE_GRAPHICS) {
         /* Kick render. */
         cmd_buffer->state.current_sub_cmd->gfx.barrier_store = true;
      }

      last_sub_cmd_type = cmd_buffer->state.current_sub_cmd->type;

      result = pvr_cmd_buffer_end_sub_cmd(cmd_buffer);
      assert(result == VK_SUCCESS);
   }

   /* Queue query program to update the buffer. */
   /* TODO NEXT: maybe do this at the next begin xfb in case the
    * counter buffers are never used again by another begin or query? */
   result = pvr_cmd_buffer_start_sub_cmd(cmd_buffer, PVR_SUB_CMD_TYPE_EVENT);
   assert(result == VK_SUCCESS);

   cmd_buffer->state.current_sub_cmd->event = (struct pvr_sub_cmd_event){
      .type = PVR_EVENT_TYPE_BARRIER,
      .barrier = {
         .wait_for_stage_mask = PVR_PIPELINE_STAGE_GEOM_BIT,
         .wait_at_stage_mask = PVR_PIPELINE_STAGE_QUERY_BIT,
      },
   };

   result = pvr_cmd_buffer_end_sub_cmd(cmd_buffer);
   assert(result == VK_SUCCESS);

   pvr_add_query_program(cmd_buffer, &query_info);

   result = pvr_cmd_buffer_start_sub_cmd(cmd_buffer, PVR_SUB_CMD_TYPE_EVENT);
   assert(result == VK_SUCCESS);

   cmd_buffer->state.current_sub_cmd->event = (struct pvr_sub_cmd_event){
      .type = PVR_EVENT_TYPE_BARRIER,
      .barrier = {
         .wait_for_stage_mask = PVR_PIPELINE_STAGE_QUERY_BIT,
         .wait_at_stage_mask = PVR_PIPELINE_STAGE_TRANSFER_BIT,
      },
   };

   result = pvr_cmd_buffer_end_sub_cmd(cmd_buffer);
   assert(result == VK_SUCCESS);

   if (last_sub_cmd_type != PVR_SUB_CMD_TYPE_INVALID) {
      result = pvr_cmd_buffer_start_sub_cmd(cmd_buffer, last_sub_cmd_type);
      assert(result == VK_SUCCESS);

      if (last_sub_cmd_type == PVR_SUB_CMD_TYPE_GRAPHICS) {
         /* Use existing render setup, but load color attachments from HW
          * BGOBJ.
          */
         cmd_buffer->state.current_sub_cmd->gfx.barrier_load = true;
         cmd_buffer->state.current_sub_cmd->gfx.barrier_store = false;
      }
   }

#if 0
   /* No Need to kick off xfb for empty program. */
   if (cmd_buffer->state.gfx_pipeline->shader_state.xfb_shader_state
          .total_varying_count) {
      pvr_end_tranform_feedback(cmd_buffer);
   }
#endif
}

/* pvr_xfb_get_counter_dwords - return the counter size in dword. */
static inline uint32_t
pvr_xfb_get_counter_dwords(const struct pvr_device *const device)
{
   const uint32_t bits = PVR_GET_FEATURE_VALUE(&device->pdevice->dev_info,
                                               transform_feedback_query_bits,
                                               32U);
   return bits == 64U ? 2 : 1;
}

/* pvr_xfb_get_counter_bytes - return the counter size in bytes. */
static inline uint32_t
pvr_xfb_get_counter_bytes(const struct pvr_device *const device)
{
   const uint32_t bits = PVR_GET_FEATURE_VALUE(&device->pdevice->dev_info,
                                               transform_feedback_query_bits,
                                               32U);
   return bits == 64U ? 8 : 4;
}

VkResult pvr_xfb_query_pool_setup(struct pvr_device *device,
                                  const VkQueryPoolCreateInfo *pCreateInfo,
                                  struct pvr_query_pool *pool)
{
   const uint32_t counter_size = pvr_xfb_get_counter_dwords(device) * sizeof(uint32_t);
   const uint32_t offsets_alloc_size = pCreateInfo->queryCount * counter_size;
   const uint32_t cache_line_size =
      rogue_get_slc_cache_line_size(&device->pdevice->dev_info);
   /* xfb queries return 2 results per query: primitives actually written and
    * attempted.
    */
   const uint32_t query_result_count = 2;
   const uint32_t offsets_accumulated_alloc_size =
      pCreateInfo->queryCount * query_result_count * counter_size * sizeof(uint32_t);
   const uint32_t result_buffer_alloc_size =
      pCreateInfo->queryCount * query_result_count * sizeof(uint64_t);
   const uint32_t offsets_failed_alloc_size =
      pCreateInfo->queryCount * counter_size * sizeof(uint32_t);
   VkResult result;

   pool->query_result_count = query_result_count;
   pool->result_stride = pool->query_size;

   result = pvr_bo_suballoc(&device->suballoc_general,
                            result_buffer_alloc_size,
                            cache_line_size,
                            false,
                            &pool->result_buffer);
   if (result != VK_SUCCESS)
      return result;

   result = pvr_bo_suballoc(&device->suballoc_general,
                            offsets_alloc_size,
                            cache_line_size,
                            false,
                            &pool->xfb_count_offsets);
   if (result != VK_SUCCESS)
      goto err_free_result_buffer;

   result = pvr_bo_suballoc(&device->suballoc_general,
                            offsets_failed_alloc_size,
                            cache_line_size,
                            false,
                            &pool->xfb_count_offsets_failed);
   if (result != VK_SUCCESS)
      goto err_free_offsets;

   result = pvr_bo_suballoc(&device->suballoc_general,
                            offsets_accumulated_alloc_size,
                            cache_line_size,
                            false,
                            &pool->xfb_count_offsets_accumulated);
   if (result != VK_SUCCESS)
      goto err_free_offsets_failed;

   return VK_SUCCESS;

err_free_offsets_failed:
   pvr_bo_suballoc_free(pool->xfb_count_offsets_failed);
err_free_offsets:
   pvr_bo_suballoc_free(pool->xfb_count_offsets);
err_free_result_buffer:
   pvr_bo_suballoc_free(pool->result_buffer);

   return result;
}

void pvr_xfb_query_pool_cleanup(const struct pvr_query_pool *pool)
{
   pvr_bo_suballoc_free(pool->xfb_count_offsets_accumulated);
   pvr_bo_suballoc_free(pool->xfb_count_offsets_failed);
   pvr_bo_suballoc_free(pool->xfb_count_offsets);
   pvr_bo_suballoc_free(pool->result_buffer);
}

uint64_t pvr_xfb_query_pool_get_result(struct pvr_device *device,
                                       struct pvr_query_pool *pool,
                                       uint32_t query_idx,
                                       uint32_t result_idx)
{
   const uint32_t query_result_idx =
      (query_idx * pool->query_result_count) + result_idx;
   uint64_t result = 0;

   if (pvr_xfb_get_counter_dwords(device) == 2U) {
      const volatile uint64_t *query_results =
         pvr_bo_suballoc_get_map_addr(pool->result_buffer);
      result = query_results[query_result_idx];
   } else {
      const volatile uint32_t *query_results =
         pvr_bo_suballoc_get_map_addr(pool->result_buffer);
      result = query_results[query_result_idx];
   }

   return result;
}

static void
pvr_xfb_query_program_input_init(struct pvr_cmd_buffer *cmd_buffer,
                                 struct pvr_query_pool *pool,
                                 uint32_t query_idx,
                                 uint32_t stream_idx,
                                 struct pvr_pds_xfb_query_program_input *input)
{
   const uint32_t counter_bytes = pvr_xfb_get_counter_bytes(cmd_buffer->device);
   struct pvr_cmd_buffer_xfb_state *xfb_state = &cmd_buffer->state.xfb_state;

   input->dev_info = &cmd_buffer->device->pdevice->dev_info;
   input->query_type = pool->query_type;
   input->counter_bytes = counter_bytes;

   if (pool->query_type == VK_QUERY_TYPE_TRANSFORM_FEEDBACK_STREAM_EXT) {
      const uint32_t counter_offset = counter_bytes * query_idx;

      input->result_addr =
         PVR_DEV_ADDR_OFFSET(pool->result_buffer->dev_addr,
                             query_idx * pool->query_result_count);
      input->availability_addr =
         PVR_DEV_ADDR_OFFSET(pool->availability_buffer->dev_addr,
                             query_idx * pool->query_size);

      input->count_offset_addr =
         PVR_DEV_ADDR_OFFSET(pool->xfb_count_offsets->dev_addr, counter_offset);
      input->count_offset_failed_addr =
         PVR_DEV_ADDR_OFFSET(pool->xfb_count_offsets_failed->dev_addr,
                             counter_offset);
      input->count_offset_accumulated_addr =
         PVR_DEV_ADDR_OFFSET(pool->xfb_count_offsets_accumulated->dev_addr,
                             counter_offset);
   }

   if (xfb_state->is_active) {
      const uint32_t counter_offset = counter_bytes * stream_idx;
      const struct pvr_graphics_pipeline *pipeline =
         cmd_buffer->state.gfx_pipeline;

      input->total_count_addr = PVR_DEV_ADDR_OFFSET(
         pipeline->shader_state.xfb_shader_state.counters->dev_addr,
         counter_offset);

      input->is_active = true;
   }
}

static void pvr_xfb_query_store_program_input_init(
   struct pvr_cmd_buffer *cmd_buffer,
   struct pvr_query_pool *pool,
   uint32_t query_idx,
   uint32_t stream_idx,
   struct pvr_pds_xfb_query_program_input *input)
{
   *input = (struct pvr_pds_xfb_query_program_input){
      .is_store = true,
   };

   pvr_xfb_query_program_input_init(cmd_buffer,
                                    pool,
                                    query_idx,
                                    stream_idx,
                                    input);
}

static void pvr_xfb_query_init_program_input_init(
   struct pvr_cmd_buffer *cmd_buffer,
   struct pvr_query_pool *pool,
   uint32_t query_idx,
   uint32_t stream_idx,
   struct pvr_pds_xfb_query_program_input *input)
{
   *input = (struct pvr_pds_xfb_query_program_input){
      .is_init = true,
   };

   pvr_xfb_query_program_input_init(cmd_buffer,
                                    pool,
                                    query_idx,
                                    stream_idx,
                                    input);
}

static inline void pvr_xfb_query_program_type_generate(
   enum pvr_pds_builder_generate_mode gen_mode,
   struct pvr_pds_xfb_query_program_input *input,
   struct pvr_pds_xfb_program_output *output)
{
   /* In preparation to have more query programs (e.g.
    * VK_QUERY_TYPE_PRIMITIVES_GENERATED_EXT), allow to select an
    * entry point here.
    */
   if (input->is_init || input->is_store) {
      pvr_pds_gen_generate_query_program(gen_mode, input, output);
   } else {
      unreachable();
   }
}

void pvr_xfb_query_pool_begin(struct pvr_cmd_buffer *cmd_buffer,
                              struct pvr_query_pool *pool,
                              uint32_t query_idx,
                              uint32_t stream_idx)
{
   struct pvr_cmd_buffer_xfb_state *xfb_state = &cmd_buffer->state.xfb_state;
   struct pvr_cmd_buffer_state *state = &cmd_buffer->state;
   uint32_t temp_size;

   if (stream_idx >= PVR_MAX_XFB_STREAMS) {
      mesa_logw("Ignoring xfb query for unsupported stream %u (max %u)",
                stream_idx,
                PVR_MAX_XFB_STREAMS);
      return;
   }

#if 0
   if (pool->query_type == VK_QUERY_TYPE_TRANSFORM_FEEDBACK_STREAM_EXT) {
      const uint32_t query_pool_index =
         PVR_QUERY_POOL_INDEX_XFB_STREAM_0 + stream_idx;

      assert(!(xfb_state->stream_query_mask & BITFIELD_BIT(stream_idx)));
      xfb_state->stream_query_mask |= BITFIELD_BIT(stream_idx);

      state->query_pool_indeces[query_pool_index] = query_idx;
      state->query_pools[query_pool_index] = pool;
   }
#endif

   assert(pool->query_type == VK_QUERY_TYPE_TRANSFORM_FEEDBACK_STREAM_EXT);

   if (state->current_sub_cmd) {
      assert(state->current_sub_cmd->type == PVR_SUB_CMD_TYPE_GRAPHICS);

      if (!state->current_sub_cmd->gfx.query_pool) {
         state->current_sub_cmd->gfx.query_pool = pool;
      } else if (state->current_sub_cmd->gfx.query_pool != pool) {
         VkResult result;

         /* Kick render. */
         state->current_sub_cmd->gfx.barrier_store = true;

         result = pvr_cmd_buffer_end_sub_cmd(cmd_buffer);
         if (result != VK_SUCCESS)
            return;

         result =
            pvr_cmd_buffer_start_sub_cmd(cmd_buffer, PVR_SUB_CMD_TYPE_GRAPHICS);
         if (result != VK_SUCCESS)
            return;

         /* Use existing render setup, but load color attachments from HW
          * BGOBJ.
          */
         state->current_sub_cmd->gfx.barrier_load = true;
         state->current_sub_cmd->gfx.barrier_store = false;
         state->current_sub_cmd->gfx.query_pool = pool;
      }
   }

   assert(!stream_idx);
   state->query_pools[PVR_QUERY_POOL_INDEX_XFB_STREAM_0] = pool;
   state->xfb_enabled = true;

   /* Add the index to the list for this render. */
   util_dynarray_append(&state->query_indices, __typeof__(query_idx), query_idx);
}

void pvr_xfb_query_pool_end(struct pvr_cmd_buffer *cmd_buffer,
                            struct pvr_query_pool *pool,
                            uint32_t query_idx,
                            uint32_t stream_idx)
{
   struct pvr_cmd_buffer_xfb_state *xfb_state = &cmd_buffer->state.xfb_state;
   struct pvr_cmd_buffer_state *state = &cmd_buffer->state;
   uint32_t temp_size;

   if (stream_idx >= PVR_MAX_XFB_STREAMS) {
      mesa_logw("Ignoring xfb query for unsupported stream %u (max %u)",
                stream_idx,
                PVR_MAX_XFB_STREAMS);
      return;
   }

#if 0
   if (pool->query_type == VK_QUERY_TYPE_TRANSFORM_FEEDBACK_STREAM_EXT) {
      const uint32_t query_pool_index =
         PVR_QUERY_POOL_INDEX_XFB_STREAM_0 + stream_idx;

      assert(xfb_state->stream_query_mask & BITFIELD_BIT(stream_idx));
      xfb_state->stream_query_mask &= ~BITFIELD_BIT(stream_idx);

      state->query_pools[query_pool_index] = NULL;
   }
#endif

   assert(pool->query_type == VK_QUERY_TYPE_TRANSFORM_FEEDBACK_STREAM_EXT);

   assert(!stream_idx);
   state->query_pools[PVR_QUERY_POOL_INDEX_XFB_STREAM_0] = NULL;
   state->xfb_enabled = false;
}

void pvr_CmdBeginQueryIndexedEXT(VkCommandBuffer commandBuffer,
                                 VkQueryPool queryPool,
                                 uint32_t query,
                                 VkQueryControlFlags flags,
                                 uint32_t index)
{
   PVR_FROM_HANDLE(pvr_cmd_buffer, cmd_buffer, commandBuffer);
   PVR_FROM_HANDLE(pvr_query_pool, pool, queryPool);

   PVR_CHECK_COMMAND_BUFFER_BUILDING_STATE(cmd_buffer);

   pvr_xfb_query_pool_begin(cmd_buffer, pool, query, index);

   assert(!flags);
   assert(!index);
   assert(pool->query_type == VK_QUERY_TYPE_TRANSFORM_FEEDBACK_STREAM_EXT);

   struct pvr_cmd_buffer_xfb_state *xfb_state = &cmd_buffer->state.xfb_state;
   xfb_state->pool = pool;
   xfb_state->query = query;
}

void pvr_CmdEndQueryIndexedEXT(VkCommandBuffer commandBuffer,
                               VkQueryPool queryPool,
                               uint32_t query,
                               uint32_t index)
{
   PVR_FROM_HANDLE(pvr_cmd_buffer, cmd_buffer, commandBuffer);
   PVR_FROM_HANDLE(pvr_query_pool, pool, queryPool);

   PVR_CHECK_COMMAND_BUFFER_BUILDING_STATE(cmd_buffer);

   pvr_xfb_query_pool_end(cmd_buffer, pool, query, index);

   assert(!index);
   assert(pool->query_type == VK_QUERY_TYPE_TRANSFORM_FEEDBACK_STREAM_EXT);

   struct pvr_cmd_buffer_xfb_state *xfb_state = &cmd_buffer->state.xfb_state;
   xfb_state->pool = NULL;
}

